%%%%%
%
% concept_learning.pl				
% Module for leaning concepts from data 
% Focus in this initial stage is complex data as images
%
% Version: 0.0.1
%
% Author: Tony Lindgren
%%%%%


% Concept learning consists of three steps:
% 1. For each class, find concepts (common patterns) for that particular class, using our "language" 
% 2. Given the learnt concepts for each class investegate if it is engough to separate the classes,
% otherwise check clases pairwise and find separating features, incl. not features,
% i.e. areas which much differ between the classes.      
% 3. formulate rules for each class (?), then  
%

:- module(concept_learning,[concept_learning/10]).
:- use_module([mat]).

concept_learning(TrainExList, _BKList, _Parameters, _Head, _RuleSetKey, _ClassIndex, _FeaturesDTL, _NewClassT, _RuleSet, _DefaultFact):-
	separate_classes(TrainExList, [], GroupedTrainClassesList),!,
%	length(GroupedTrainClassesList, NoClasses),
	find_concepts(GroupedTrainClassesList, ConceptsLists),trace,
	write(ConceptsLists).


find_concepts(ClassesList, ConceptsLists):-
%	write('Class: '),write(Class),nl,
	find_bounding_boxes(ClassesList, 0-0-0-0, Top-Left-Bottom-Right, 0, TotCount, ExWBB),
	MeanVerticalDiff is round((Right-Left) / TotCount),
	MeanHorizontalDiff is round((Bottom-Top) / TotCount),
	write('MeanHorizontalDiff: '),write(MeanHorizontalDiff),nl,
	write('MeanVerticalDiff:'), write(MeanVerticalDiff),nl,
	Tot is MeanHorizontalDiff + MeanVerticalDiff,
	MeanDiff is round(Tot/2), 
	write('MeanDiff:'), write(MeanDiff),nl,
	transform_exs_to_mean(ExWBB, ExTransformed),
	find_common_patterns(ExTransformed, ConceptsLists).

find_common_patterns([], []):-!.
find_common_patterns([concept_info(Class,CTop,CLeft,CBottom,CRight,TotCount,[First|Rest])|CInfo],
		     [concept_info(Class,CTop,CLeft,CBottom,CRight,TotCount,FullMHR,FullMNHR,TransMHR,TransMNHR,[First|Rest])|CAddInfo]):-
	First = mat_ex(Ex, _, ExTransMat, _, _, _, _),
	mat_copy_mat(Ex, FullMHR),
	mat_copy_mat(Ex, FullMNHR),
	mat_copy_mat(ExTransMat, TransMHR),
	mat_copy_mat(ExTransMat, TransMNHR),
	find_common_class_patterns(Rest, TotCount, FullMHR, FullMNHR, TransMHR, TransMNHR),
	write('Class:'),write(Class),nl,
	mat_pp(FullMHR),
	mat_pp(FullMNHR),
	mat_pp(TransMHR),
	mat_pp(TransMNHR),
	!,
	find_common_patterns(CInfo, CAddInfo).

find_common_class_patterns([], TotCount, FullMHR, _, TransMHR, _):-
	MultFactor is 1/TotCount,
	mat_mult(FullMHR,MultFactor),
	mat_mult(TransMHR,MultFactor),!.
find_common_class_patterns([mat_ex(Ex, _, ExTransMat, _, _, _, _)|Rest], TotCount, FullMHR, FullMNHR, TransMHR, TransMNHR):-
	mat_1_pair(Ex, FullMHR),
	mat_0_pair(Ex, FullMNHR), 
	mat_1_pair(ExTransMat, TransMHR),
	mat_0_pair(ExTransMat, TransMNHR),!,
	find_common_class_patterns(Rest, TotCount, FullMHR, FullMNHR, TransMHR, TransMNHR).
	
transform_exs_to_mean([], []):-!.
transform_exs_to_mean([concept_info(Class,CTop,CLeft,CBottom,CRight,PartTotCount,ExsWBB)|CInfo],
		      [concept_info(Class,CTop,CLeft,CBottom,CRight,PartTotCount,TransformedExs)|CAddInfo]):-
	RowSize is round((CBottom-CTop)/PartTotCount),
	ColSize is round((CRight-CLeft)/PartTotCount),
	transform_class_exs_to_mean(ExsWBB, RowSize, ColSize, TransformedExs),!,
	transform_exs_to_mean(CInfo, CAddInfo).

transform_class_exs_to_mean([], _, _, []).
transform_class_exs_to_mean([mat_ex(Ex, Top, Left, Bottom, Right)|Exs], RowSize, ColSize,
			    [mat_ex(Ex, ExBBRegion, ExTransMat, Top, Left, Bottom, Right)|TransExs]):-
	%mat_pp(Ex),
	mat_sub_region(Ex, Top, Left, Bottom, Right, ExBBRegion), 
	mat(ExTransMat, RowSize, ColSize),
	%mat_pp(ExBBRegion),
	mat_project_into(ExBBRegion, ExTransMat),!,
	%mat_pp(ExTransMat),
	%trace,
	transform_class_exs_to_mean(Exs, RowSize, ColSize, TransExs).	
	 	
find_bounding_boxes([], Top-Left-Bottom-Right, Top-Left-Bottom-Right, TotCount, TotCount, []):-!.
find_bounding_boxes([Class-ExList|ClassExList], TempT-TempL-TempB-TempR, FTop-FLeft-FBottom-FRight, TempTC, FTotCount,
		    [concept_info(Class,CTop,CLeft,CBottom,CRight,PartTotCount,ExWBB)|ConceptsList]):-
	write('Class: '),write(Class),nl,
	%length(ExList),
	find_class_bounding_boxes(ExList, ExWBB, 0-0-0-0, CTop-CLeft-CBottom-CRight, 0, PartTotCount),
	NewTempT is TempT + CTop,
	NewTempL is TempL + CLeft,
	NewTempB is TempB + CBottom,
	NewTempR is TempR + CRight,
	length(ExList, PartTempTC),
	NewTempTC is TempTC +  PartTempTC,!,
	%use this values as guidelines to do projection to a mean (equal) surface only using bits (i.e. 0 = 0, 1 or grater is 1)
        %do more stuff here
	%trace,
	find_bounding_boxes(ClassExList, NewTempT-NewTempL-NewTempB-NewTempR, FTop-FLeft-FBottom-FRight, NewTempTC, FTotCount, ConceptsList).
find_bounding_boxes(_, _A, _B, _C, _D, _E):-
	write('Error'),nl,
	trace.	   

find_class_bounding_boxes([], [], Top-Left-Bottom-Right, Top-Left-Bottom-Right, TotCount, TotCount):-!.
find_class_bounding_boxes([Ex|Exs], [mat_ex(Ex, Top, Left, Bottom, Right)|MatExes],
			  TempT-TempL-TempB-TempR, FTop-FLeft-FBottom-FRight, TempTC, FTotCount):-
	%mat_pp(Ex),
	find_bounding_box(Ex, Top, Left, Bottom, Right),!,
	NewTempT is TempT + Top,
	NewTempL is TempL + Left,
	NewTempB is TempB + Bottom,
	NewTempR is TempR + Right,
	NewTempTC is TempTC + 1,
	%write('Top: '), write(Top), nl,
	%write('Left: '), write(Left), nl,
 	%write('Bottom: '), write(Bottom),nl,
	%write('Right: '), write(Right),nl,
	find_class_bounding_boxes(Exs, MatExes, NewTempT-NewTempL-NewTempB-NewTempR, FTop-FLeft-FBottom-FRight, NewTempTC, FTotCount).

%assumes background = 0
find_bounding_box(Ex, Top, Left, Bottom, Right):-
        mat_no_rows(Ex, NoRows),
	mat_no_cols(Ex, NoCols),
	find_top(Ex, 1, NoRows, Top),
	find_left(Ex, 1, NoCols, Left),
	find_bottom(Ex, NoRows, Bottom),
	find_right(Ex, NoCols, Right).

find_top(_, NoRows, NoRows, none):-
	write('Did not find top!').
find_top(Ex, CurrRow, NoRows, Top):-
	mat_sum_row(Ex, CurrRow, Sum),
	(Sum > 0 ->
	  Top = CurrRow
	;
	  NewCurrRow is CurrRow + 1,
	  find_top(Ex, NewCurrRow, NoRows, Top)
	).

find_left(_, NoCols, NoCols, none):-
	write('Did not find left!').
find_left(Ex, CurrCol, NoCols, Left):-
	mat_sum_col(Ex, CurrCol, Sum),
	(Sum > 0 ->
	  Left = CurrCol
	;
	  NewCurrCol is CurrCol + 1,
	  find_left(Ex, NewCurrCol, NoCols, Left)
	).

find_bottom(_, 0, none):-
	write('Did not find bottom!').
find_bottom(Ex, CurrRow, Bottom):-
	mat_sum_row(Ex, CurrRow, Sum),
	(Sum > 0 ->
	  Bottom = CurrRow
	;
	  NewCurrRow is CurrRow - 1,
	  find_bottom(Ex, NewCurrRow, Bottom)
	).

find_right(_, 0, none):-
	write('Did not find right!').
find_right(Ex, CurrCol, Right):-
	mat_sum_col(Ex, CurrCol, Sum),
	(Sum > 0 ->
	  Right = CurrCol
	;
	  NewCurrCol is CurrCol - 1,
	  find_right(Ex, NewCurrCol, Right)
	).

separate_classes([], GTCL, GTCL).
separate_classes([mat_ex(MatEx, ClassVal)|ExTrainMatList], TempGTCL, FinalGTCL):-
	(select(ClassVal-MatExList, TempGTCL, RestGTCL) ->
	  append([MatEx], MatExList, NewMatExList),
	  select(ClassVal-NewMatExList, NewTempGTCL, RestGTCL),
	  separate_classes(ExTrainMatList, NewTempGTCL, FinalGTCL)
	;
	  append([ClassVal-[MatEx]], TempGTCL, NewTempGTCL),
  	  separate_classes(ExTrainMatList, NewTempGTCL, FinalGTCL)
	).