%%%%%
%
% do_induction.pl				
% Module that facilitate usage of various induction algorithms, handle multi-threading etc.   
%
% Version: 0.0.1
%
% Author: Tony Lindgren
%%%%%


:- module(do_induction,[do_induction/12]).

:- use_module([util,  tree, rule_set, handle_rules, concept_learning]).

do_induction(ensemble_model(Size, Parameters, dac), Head, TrainExLists, BKList, BBKey, ClassIndex, FeaturesDTL, BBKeys, Trees, NoTrees, ClassT, Rules):-
    join_folds(TrainExLists, NewTrainExList),!,
    (member(mt(yes),Parameters) ->
      do_trees_mt(Size, Parameters, Head, NewTrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, ClassT, BBKeys, Trees, NoTrees, Rules)
    ;
      do_trees(Size, Parameters, Head, NewTrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, ClassT, BBKeys, Trees, NoTrees, Rules)
    ).
do_induction(single_model(Parameters, dac), Head, TrainExLists, BKList, BBKey, ClassIndex, FeaturesDTL, BBKeys, Trees, NoTrees, ClassT, Rules):-
    join_folds(TrainExLists, NewTrainExList),!,
    do_trees(1, Parameters, Head, NewTrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, ClassT, BBKeys, Trees, NoTrees, Rules).

do_induction(ensemble_model(Size, Parameters, sac), Head, TrainExLists, BKList, BBKey, ClassIndex, FeaturesDTL, BBKeys, Trees, NoTrees, ClassT, Rules):-
    join_folds(TrainExLists, NewTrainExList),!,
    (member(mt(yes),Parameters) ->
      do_rule_set_mt(Size, Parameters, Head, NewTrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, ClassT, BBKeys, Trees, NoTrees, Rules)
    ;
      do_rule_set(Size, Parameters, Head, NewTrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, ClassT, BBKeys, Trees, NoTrees, Rules)
    ).
do_induction(single_model(Parameters, sac), Head, TrainExLists, BKList, BBKey, ClassIndex, FeaturesDTL, BBKeys, Trees, NoTrees, ClassT, Rules):-
    join_folds(TrainExLists, NewTrainExList),!,
    do_rule_set(1, Parameters, Head, NewTrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, ClassT, BBKeys, Trees, NoTrees, Rules).

do_induction(single_model(Parameters, concept_learning), Head, TrainExLists, BKList, BBKey, ClassIndex, FeaturesDTL, BBKeys, Trees, NoTrees, ClassT, Rules):-
    join_folds(TrainExLists, NewTrainExList),!,
    do_concept_learning(1, Parameters, Head, NewTrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, ClassT, BBKeys, Trees, NoTrees, Rules).

%%
% join the different folds to one list
join_folds([], []).
join_folds([Fold|Folds], TrainExList):-
   join_folds(Folds, PartList),!,
   append(Fold, PartList, TrainExList).

do_rule_set_mt(Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, FCT, RuleSetKeys, RuleSets, NoRules, Rules):-
    (current_thread_pool(ensPool) ->
      do_c_rule_sets_mt(ensPool, ensQueue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, FCT, RuleSetKeys, RuleSets, NoRules, Rules)
    ; 
      current_prolog_flag(cpu_count, NoCores),
      NoCoresForPool is NoCores - 1,
      message_queue_create(ensQueue),
      thread_pool_create(ensPool, NoCoresForPool, []),
      do_c_rule_sets_mt(ensPool, ensQueue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, FCT, RuleSetKeys, RuleSets, NoRules, Rules)
    ).

do_c_rule_sets_mt(Pool, Queue, 0, Parameters, _, _, _, _, ClassIndex, _, _, FCT, PTreeKey, PTrees, PNoTrees, PRules):-
   thread_pool_property(Pool, running(Members)),
   collect_all_rule_set_result(Members, Queue, Parameters, PTreeKey, PTrees, _, FCT, PNoTrees, PRules, ClassIndex).
do_c_rule_sets_mt(Pool, Queue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, FCT, FRuleSetsKeys, FRuleSets, FNoRuleSets, FRules):-
   Size > 0,	
   thread_pool_property(Pool, free(No)),
   (No > 0 ->
     do_c_rule_set_thread(Pool, Queue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL), 
     NewSize is Size - 1,
     do_c_rule_sets_mt(Pool, Queue, NewSize, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, FCT, FRuleSetsKeys, FRuleSets, FNoRuleSets, FRules)
   ;
     collect_rule_set_result(Queue, Parameters, PRuleSetKey, PRuleSet, NewCT, PNoRulesSet, PRules, ClassIndex),
     do_c_rule_sets_mt(Pool, Queue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, NewCT, FCT, RRuleSetKey, RRuleSet, RNoRuleSet, RRules),
     append([PRuleSetKey], RRuleSetKey, FRuleSetsKeys),
     append([PRuleSet], RRuleSet, FRuleSets),
     append([PNoRulesSet], RNoRuleSet, FNoRuleSets),
     append([rule_set(PRules)], RRules, FRules)
   ). 

do_c_rule_set_thread(Pool, Queue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL):- 
   (member(bagging, Parameters) ->
         create_bag(TrainExList, Size, NewTrainExList)
    ;
         NewTrainExList = TrainExList 
    ),
    write_to_codes(Size-BBKey, RuleSetCodes),
    atom_codes(RuleSetKey, RuleSetCodes),
    thread_create_in_pool(Pool, rule_set_m_t(Queue, NewTrainExList, BKList, Parameters, Head, RuleSetKey, ClassIndex, FeaturesDTL), _, []).

rule_set_m_t(Queue, TrainExList, BKList, Parameters, Head, BBKey, ClassIndex, FeaturesDTL):-
    rule_set(TrainExList, BKList, Parameters, Head, BBKey, ClassIndex, FeaturesDTL, ClassT, RuleSet, _),
    thread_self(ID),
    thread_send_message(Queue, result(ClassT, Head, BBKey, ClassIndex, RuleSet, ID)).

collect_all_rule_set_result(0, _, _, [], [], FCT, FCT, [], [], _).
collect_all_rule_set_result(MessagesLeft, QueueId, Parameters, FTreeKeys, FTrees, _, FCT, FNoTrees, FRules, ClassIndex):-
	collect_rule_set_result(QueueId, Parameters, PTreeKey, PTrees, NewCT, PNoTrees, PRules, ClassIndex),
	NewMessagesLeft is MessagesLeft - 1,
	collect_all_rule_set_result(NewMessagesLeft, QueueId, Parameters, RTreeKeys, RTrees, NewCT, FCT, RNoTrees, RRules, ClassIndex),
	append([PTreeKey], RTreeKeys, FTreeKeys),
        append([PTrees], RTrees, FTrees),
        append([PNoTrees], RNoTrees, FNoTrees),
        append([rule_set(PRules)], RRules, FRules).

collect_rule_set_result(QueueId, Parameters, RuleSetKey, RuleSet, NewCT, NoRules, Rules, ClassIndex):- 
    thread_get_message(QueueId, result(NewCT, Head, RuleSetKey, ClassIndex, RuleSet, ID)),
    (member(bb_rules, Parameters) ->
      assert_rule_set_to_bb(RuleSet, Head, RuleSetKey, 0, NoRules, Rules)
    ;
      Head =.. [_|HVarList],
      rough_rule_set_to_rules(RuleSet, HVarList, 0, NoRules, Rules)
    ),
    thread_join(ID,_).

%use ordinary rule set instead...?
do_concept_learning(0, _Parameters, _Head, _TrainExList, _BKList, _BKKey, _ClassIndex, _FeaturesDTL, ClassT, ClassT, [], [], [], []).
do_concept_learning(Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _ClassTuple, FCT, [RuleSetKey|Keys], [RuleSet|RuleSets], [NoRules|Rest], [rule_set(Rules)|RRules]):-
    (member(bagging, Parameters) ->
         create_bag(TrainExList, Size, NewTrainExList)
    ;
         NewTrainExList = TrainExList
    ),
    %write('RuleSet: '), write(Size),nl,
    write_to_codes(Size-BBKey, RuleSetCodes),
    atom_codes(RuleSetKey, RuleSetCodes),
    %trace,
    concept_learning(NewTrainExList, BKList, Parameters, Head, RuleSetKey, ClassIndex, FeaturesDTL, NewClassT, RuleSet, _DefaultFact),
    NewSize is Size - 1,!,    
    (member(bb_rules, Parameters) ->
      assert_rule_set_to_bb(RuleSet, Head, RuleSetKey, 0, NoRules, Rules),
      do_concept_learning(NewSize, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, NewClassT, FCT, Keys, RuleSets, Rest, RRules)
    ;
      Head =.. [_|HVarList],
      rough_rule_set_to_rules(RuleSet, HVarList, 0, NoRules, Rules),   
      do_concept_learning(NewSize, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, NewClassT, FCT, Keys, RuleSets, Rest, RRules)
    ).

do_rule_set(0, _Parameters, _Head, _TrainExList, _BKList, _BKKey, _ClassIndex, _FeaturesDTL, ClassT, ClassT, [], [], [], []).
do_rule_set(Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _ClassTuple, FCT, [RuleSetKey|Keys], [RuleSet|RuleSets], [NoRules|Rest], [rule_set(Rules)|RRules]):-
    (member(bagging, Parameters) ->
         create_bag(TrainExList, Size, NewTrainExList)
    ;
         NewTrainExList = TrainExList
    ),
    write_to_codes(Size-BBKey, RuleSetCodes),
    atom_codes(RuleSetKey, RuleSetCodes),
    rule_set(NewTrainExList, BKList, Parameters, Head, RuleSetKey, ClassIndex, FeaturesDTL, NewClassT, RuleSet, _DefaultFact),
    NewSize is Size - 1,!,    
    (member(bb_rules, Parameters) ->
      assert_rule_set_to_bb(RuleSet, Head, RuleSetKey, 0, NoRules, Rules), %find this predicate (probably old code)
      do_rule_set(NewSize, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, NewClassT, FCT, Keys, RuleSets, Rest, RRules)
    ;
      Head =.. [_|HVarList],
      rough_rule_set_to_rules(RuleSet, HVarList, 0, NoRules, Rules),   
      do_rule_set(NewSize, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, NewClassT, FCT, Keys, RuleSets, Rest, RRules)
    ).

do_trees_mt(Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, FCT, TreeKeys, Trees, NoTrees, Rules):-
    (current_thread_pool(ensPool) ->
      do_c_trees_mt(ensPool, ensQueue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, FCT, TreeKeys, Trees, NoTrees, Rules)
    ;
      current_prolog_flag(cpu_count, NoCores),
      NoCoresForPool is NoCores - 1,
      message_queue_create(ensQueue),
      thread_pool_create(ensPool, NoCoresForPool, []),
      do_c_trees_mt(ensPool, ensQueue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, FCT, TreeKeys, Trees, NoTrees, Rules)
    ).

do_c_trees_mt(Pool, Queue, 0, Parameters, _, _, _, _, ClassIndex, _, _, FCT, PTreeKey, PTrees, PNoTrees, PRules):-
   thread_pool_property(Pool, running(Members)),
   collect_all_tree_result(Members, Queue, Parameters, PTreeKey, PTrees, _, FCT, PNoTrees, PRules, ClassIndex).
do_c_trees_mt(Pool, Queue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, FCT, FTreeKeys, FTrees, FNoTrees, FRules):-
   Size > 0,	
   thread_pool_property(Pool, free(No)),
   (No > 0 ->
     do_c_tree_thread(Pool, Queue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL), 
     NewSize is Size - 1,
     do_c_trees_mt(Pool, Queue, NewSize, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _, FCT, FTreeKeys, FTrees, FNoTrees, FRules)
   ;
     collect_tree_result(Queue, Parameters, PTreeKey, PTrees, NewCT, PNoTrees, PRules, ClassIndex),
     do_c_trees_mt(Pool, Queue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, NewCT, FCT, RTreeKeys, RTrees, RNoTrees, RRules),
     append([PTreeKey], RTreeKeys, FTreeKeys),
     append([PTrees], RTrees, FTrees),
     append([PNoTrees], RNoTrees, FNoTrees),
     append([tree(PRules)], RRules, FRules)
   ).
   
do_c_tree_thread(Pool, Queue, Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL):- %, _, NewCT, TreeKey, Trees, NoTrees, Rules):-	
   (member(bagging, Parameters) ->
         create_bag(TrainExList, Size, NewTrainExList)
    ;
         NewTrainExList = TrainExList
    ),
    write_to_codes(Size-BBKey, TreeCodes),
    atom_codes(TreeKey, TreeCodes),
    thread_create_in_pool(Pool, tree_m_t(Queue, NewTrainExList, BKList, Parameters, Head, TreeKey, ClassIndex, FeaturesDTL), _, []).

tree_m_t(Queue, TrainExList, BKList, Parameters, Head, BBKey, ClassIndex, FeaturesDTL):-
    tree(TrainExList, BKList, Parameters, Head, BBKey, ClassIndex, FeaturesDTL, ClassT, Tree, _),	
    thread_self(ID),
    thread_send_message(Queue, result(ClassT, BBKey, ClassIndex, Tree, ID)).

collect_all_tree_result(0, _, _, [], [], FCT, FCT, [], [], _).
collect_all_tree_result(MessagesLeft, QueueId, Parameters, FTreeKeys, FTrees, _, FCT, FNoTrees, FRules, ClassIndex):-
	collect_tree_result(QueueId, Parameters, PTreeKey, PTrees, NewCT, PNoTrees, PRules, ClassIndex),
	NewMessagesLeft is MessagesLeft - 1,
	collect_all_tree_result(NewMessagesLeft, QueueId, Parameters, RTreeKeys, RTrees, NewCT, FCT, RNoTrees, RRules, ClassIndex),
	append([PTreeKey], RTreeKeys, FTreeKeys),
        append([PTrees], RTrees, FTrees),
        append([PNoTrees], RNoTrees, FNoTrees),
        append([tree(PRules)], RRules, FRules).

collect_tree_result(QueueId, Parameters, TreeKey, Tree, NewCT, NoTrees, Rules, ClassIndex):-
    thread_get_message(QueueId, result(NewCT, TreeKey, ClassIndex, Tree, ID)),
    interpret_results(Parameters, Tree, TreeKey, NoTrees, Rules),
    thread_join(ID,_).

interpret_results(Parameters, Tree, TreeKey, NoTrees, Rules):-
    (member(bb_rules, Parameters) ->
      assert_tree_to_bb(Tree, TreeKey, NoTrees, Rules)      
    ;
      tree_to_rules(Tree, NoTrees, Rules)
    ).

do_trees(0, _Parameters, _Head, _TrainExList, _BKList, _BKKey, _ClassIndex, _FeaturesDTL, ClassT, ClassT, [], [], [], []).
do_trees(Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, _ClassTuple, FCT, [TreeKey|Keys], [Tree|Trees],[NoTrees|Rest], [tree(Rules)|RRules]):-
   %((Size = 1, BBkey = 7 ) -> trace;true),	
    %write('Tree: '), write(Size),nl,
    (member(bagging, Parameters) ->
         %write('Creating bag...'),nl,
         create_bag(TrainExList, Size, NewTrainExList)
         %write('Bag complete...'),nl
    ;
         NewTrainExList = TrainExList
    ),
    write_to_codes(Size-BBKey, TreeCodes),
    atom_codes(TreeKey,TreeCodes),
    tree(NewTrainExList, BKList, Parameters, Head, TreeKey, ClassIndex, FeaturesDTL, NewClassT, Tree, _DefaultFact),
    NewSize is Size - 1,!,    
    (member(bb_rules, Parameters) ->
      assert_tree_to_bb(Tree, TreeKey, NoTrees, Rules),
      do_trees(NewSize, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, NewClassT, FCT, Keys, Trees, Rest, RRules)
    ;
      tree_to_rules(Tree, NoTrees, Rules),
      do_trees(NewSize, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, NewClassT, FCT, Keys, Trees, Rest, RRules)
    ).

do_trees(Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, ClassTuple, FCT, [TreeKey|Keys], [Tree|Trees], [NoTrees|Rest], [tree(Rules)|RRules]):-  
    do_trees(Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, FeaturesDTL, ClassTuple, FCT, [TreeKey|Keys], [Tree|Trees], [NoTrees|Rest], [tree(Rules)|RRules]).


%%
% random sampling for bagging
create_bag(TrainExList, Size, NewTrainExList):-
    length(TrainExList, NoExamples),!,
    create_bag(TrainExList, Size, NoExamples, NoExamples, NewTrainExList).
create_bag(TrainExList, Size, NewTrainExList):-
    write('Failed creating bag'),nl,
    length(TrainExList, NoExamples),
    create_bag(TrainExList, Size, NoExamples, NoExamples, NewTrainExList).
create_bag(_, _, 0, _, []).
create_bag(TrainExList, FSize, NoExLeft, Size, [Example|Bag]):-
    random(0, Size, Rand),
    nth0(Rand, TrainExList, Example),
    NewNoExLeft is NoExLeft - 1,!,
    create_bag(TrainExList, FSize, NewNoExLeft, Size, Bag).
create_bag(TrainExList, FSize, NoExLeft, Size, Bag):-
    write('Failed creating random bag ex'), %try again
    create_bag(TrainExList, FSize, NoExLeft, Size, Bag).

rough_rule_set_to_rules([], _, NoRules, NoRules, []).
rough_rule_set_to_rules([rule(RoughCondList, Dist)|Rules], HVarList, TNoRules, FNoRules, [rule(IndexedCondList, Dist)|Rest]):-
    collect_rough_conditions(RoughCondList, HVarList, IndexedCondList),
    NewTNoRules is TNoRules + 1,
    rough_rule_set_to_rules(Rules, HVarList, NewTNoRules, FNoRules, Rest).

collect_rough_conditions([], _, []).
collect_rough_conditions([categoric(Var,_)=Val|RoughCondList], HVarList, [index(Index) = Val|Rest]):-   	
    nth1(Index, HVarList, CondVar),
    CondVar == Var,!,
    collect_rough_conditions(RoughCondList, HVarList, Rest).
collect_rough_conditions([numeric(Var)<Val|RoughCondList], HVarList, [index(Index) < Val|Rest]):-   	
    nth1(Index, HVarList, CondVar),
    CondVar == Var,!,
    collect_rough_conditions(RoughCondList, HVarList, Rest).
collect_rough_conditions([numeric(Var)>=Val|RoughCondList], HVarList, [index(Index) >= Val|Rest]):-   	
    nth1(Index, HVarList, CondVar),
    CondVar == Var,!,
    collect_rough_conditions(RoughCondList, HVarList, Rest).