%%%%%
%
% experiment_method_data.pl				
% Module that for describing type of experiment, methods to use and data set to run 
%
% Version: 0.0.1
%
% Author: Tony Lindgren
%%%%%

:- module(experiment_method_data,[experiment/1, method/1, data/1, start/0]).

:- use_module([pmls]).

setup:-
    working_directory(L,L),
    working_directory(L,'/tony/ijmlc/source/data/'). 

start:-
  start_experiment.

%Define how the experiment should be conducted
experiment([seed(42), validation_size(0.0), fold_x_v(10)]).

%Define what methods to use and their parameters
method([
%Random Forest 200
        ensemble_model(200, [classification, mt(yes), bagging, wv(no), rnd_feature(yes, 0.4), min_cov(10), min_margin(0.7), rel_margin(0)], dac),
%Random Rules - feature and coverage      
	ensemble_model(200, [classification, mt(yes), rand_rules(0.2), wv(no), rnd_feature(yes, 0.4), min_cov(10), min_margin(0.7), rel_margin(0)], sac), 
%Random Rules - bagging and feature
        ensemble_model(200, [classification, mt(yes), bagging, wv(no), rnd_feature(yes, 0.4), min_cov(10), min_margin(0.7), rel_margin(0)], sac),  
%Random Rules - bagging, feature and coverage      
	ensemble_model(200, [classification, mt(yes), bagging, rand_rules(0.2), wv(no), rnd_feature(yes, 0.4), min_cov(10), min_margin(0.7), rel_margin(0)], sac)
       ]).

%Select which data set to run the experiment on			
data([ 
       %set(acute_diagnosis_1,  csv, [filename('csv_data/mixed_data/acute_diagnosis_ex_1.csv')]),
       %set(acute_diagnosis_2,  csv, [filename('csv_data/mixed_data/acute_diagnosis_ex_2.csv')]),
       %set(adult,  csv, [filename('csv_data/mixed_data/adult_ex.csv')]),
       %set(arrhythmia,  csv, [filename('csv_data/mixed_data/arrhythmia_ex.csv')]),
       %set(balance_scale,  csv, [filename('csv_data/categorical_data/balance_scale_ex.csv')]), 
       %set(breast_cancer_wisconsin, prolog, [background_head(relation/10),   background_file('prolog_data/numerical_data/breast_cancer_wisconsin_cx'), example_file('prolog_data/numerical_data/breast_cancer_wisconsin_ex_mod')]), 
       %set(bupa,                    prolog, [background_head(bupa/7),        background_file('prolog_data/numerical_data/bupa'),                       example_file('prolog_data/numerical_data/bupa_ex')]),
       %set(cleveland_heart_disease, prolog, [background_head(relation/14),   background_file('prolog_data/numerical_data/cleveland_heart_disease_cx'), example_file('prolog_data/numerical_data/cleveland_heart_disease_ex_mod')]),
       %set(climate_model_failures,  csv, [filename('csv_data/numerical_data/climate_model_failures.csv')]),
       %set(crx,  csv, [filename('csv_data/mixed_data/crx_ex.csv')]),
       %set(forest_types, csv, [filename('csv_data/numerical_data/forest_types_ex.csv')]),
       %set(glass,                   prolog, [background_head(glass/10),      background_file('prolog_data/numerical_data/glass'),                      example_file('prolog_data/numerical_data/glass_ex')]),
       %set(haberman,                prolog, [background_head(haberman/4),    background_file('prolog_data/numerical_data/haberman'),                   example_file('prolog_data/numerical_data/haberman_data')]),
       set(iris,  csv, [filename('csv_data/numerical_data/iris_data.csv')])
       %set(image_segmentation,      prolog, [background_head(relation/20),   background_file('prolog_data/numerical_data/image_segmentation_cx'),      example_file('prolog_data/numerical_data/image_segmentation_ex')]),
       %set(ionosphere,              prolog, [background_head(good_radar/35), background_file('prolog_data/numerical_data/ionosphere'),                 example_file('prolog_data/numerical_data/ionosphere_ex')]),
       %set(magic04, csv, [filename('csv_data/numerical_data/magic04_ex.csv')]),
       %set(mushroom, csv, [filename('csv_data/categorical_data/mushroom_ex.csv')]),
       %set(pendigitis,              prolog, [background_head(rel/17),        background_file('prolog_data/numerical_data/pendigits_names'),            example_file('prolog_data/numerical_data/pendigits_ex')]),
       %set(pima_indians,            prolog, [background_head(diabetes/9),    background_file('prolog_data/numerical_data/pima_indians'),               example_file('prolog_data/numerical_data/pima_indians_ex')]), 
       %set(sensor_readings_24, csv, [filename('csv_data/numerical_data/sensor_readings_24_ex.csv')]),
       %set(shuttle,  csv, [filename('csv_data/categorical_data/shuttle_ex.csv')]), 
       %set(sonar,                   prolog, [background_head(mine/61),       background_file('prolog_data/numerical_data/sonar'),                     example_file('prolog_data/numerical_data/sonar_ex')]),
       %set(spambase, csv, [filename('csv_data/numerical_data/spambase_ex.csv')]),
       %set(spectf,                  prolog, [background_head(relation/45),   background_file('prolog_data/numerical_data/spectf_cx'),                  example_file('prolog_data/numerical_data/spectf_ex')]), 
       %set(thyroid,                 prolog, [background_head(thyroid/6),     background_file('prolog_data/numerical_data/new_thyroid'),                example_file('prolog_data/numerical_data/new_thyroid_ex')]), 
       %set(transfusion, csv, [filename('csv_data/numerical_data/transfusion_ex.csv')]), 
       %set(wine,                    prolog, [background_head(relation/14),   background_file('prolog_data/numerical_data/wine_cx'),                    example_file('prolog_data/numerical_data/wine_ex')])
     ]).

:- setup.