%%%
% Post process rules
%
% Do various post-processing issues incl. classifying a test set
%
% Author: Tony Lindgren 
%
% version 0.0.1
%
%

:- module(post_process_rules,[post_process_rules/14]).

:- use_module([library(lists), library(random), util, indexing]).

post_process_rules(DataName, Head, TestFold, Method, TestExList, TrainExList, BKList, ClassIndex, Keys, NoRules, NoConds, ClassT, Rules, Result):-
     get_parameters(Method, Parameters),
     member(generate_ex(List), Parameters),!,       % use sampling methods
     ensemble_model(_, _, Type) = Method,           % must be an ensemble or else no use in sampling!!
     copy_term(ClassT, CClassTuples),
     join_folds(TrainExList, NewTrainExList),
     use_sampling_methods(List, Rules, DataName, NewTrainExList, BKList, TestExList, TestFold, Type, Parameters, Head, ClassIndex, CClassTuples, SampRes),
     classify_rules(DataName, Method, TestExList, ClassIndex, CClassTuples, Keys, NoRules, NoConds, Rules, PartResult),     
     append(PartResult, SampRes, Result).
post_process_rules(DataName, _Head, TestFold, Method, TestExList, TrainExList, _BKList, ClassIndex, Keys, NoRules, NoConds, ClassT, Rules, Result):-
     get_parameters(Method, Parameters),
     member(index(IParams), Parameters),!,          % use indexing
     copy_term(ClassT, CClassTuples),
     join_folds(TrainExList, NewTrainExList),
     [Ex|_] = NewTrainExList,
     functor(Ex, _, Arity),
     att_min_max(NewTrainExList, Arity, ClassIndex, AttMinMaxList),
     member(bit_size(BitSizeList), IParams),
     add_ex_id(TestExList, TestExWIdList),         %for debugging as of now    
     use_indexing1(BitSizeList, TestFold, Rules, AttMinMaxList, TestExWIdList, DataName, IParams, ClassIndex, CClassTuples, IndexRes),
     classify_rules(DataName, Method, TestExWIdList, ClassIndex, CClassTuples, Keys, NoRules, NoConds, Rules, PartResult),
     append(PartResult, IndexRes, Result).
post_process_rules(DataName, _Head, _TestFold, Method,  TestExList, _TrainExList, _BKList, ClassIndex, Keys, NoRules, NoConds, ClassT, Rules, Result):-
    %(TestFold = 9 ->trace;true),	
     %write('No post-processing'),nl,%trace,
     copy_term(ClassT, CClassTuples),
     %write('Before classifying rules'),nl,%trace,
     classify_rules(DataName, Method, TestExList, ClassIndex, CClassTuples,  Keys, NoRules, NoConds, Rules, Result).%,
     %write('After clasifying rules'),nl.

classify_rules(DataName, Method, TestExList, ClassIndex, ClassTuples, Keys, NoRules, NoConds, _Rules,
	       [result(DataName, MethodName, Size, Keys, ClassTuples, NoRules, NoConds, PartEval, EnsEvalTime)]):-
       ((Method = ensemble_model(Size, Settings, dac),
	 MethodName = ensemble_dac_bb_rules
	)
       ;
       (Method = single_model(Settings, dac),
	 MethodName = single_dac_bb_rules,
	 Size = 1
        )
       ),
       member(bb_rules, Settings),!,
       %TODO:add setting to classify_test_set to cover wv(yes) case!!!
       %collect_conditions(Rules, ClassIndex, RulesList),
       statistics(runtime, _),
       classify_test_set(dac, TestExList, ClassIndex, ClassTuples, ClassTuples, ClassDist, Keys, Res),
       statistics(runtime, [_,EnsEvalTime]),
       %check_ties(Res, NoTies),
       %write('We found this amount of ties: ' ), write(NoTies),nl,
       length(Res, ResL),
       evaluate(ClassDist, ClassTuples, Res, ResL, PartEval).
classify_rules(DataName, Method, TestExList, ClassIndex, ClassTuples, Keys, NoRules, NoConds, Rules,
	       [result(DataName, MethodName, Size, Keys, ClassTuples, NoRules, NoConds, PartEval, EnsEvalTime)]):-
       ((Method = ensemble_model(Size, Settings, dac),
	 MethodName = ensemble_dac_list_of_rules
	)
       ;
       (Method = single_model(Settings, dac),
	 MethodName = single_dac_list_of_rules,
	 Size = 1
        )
       ),!,
       %trace,
       collect_conditions(Rules, ClassIndex, RulesList),
       statistics(runtime, _),
       %trace,Rules -> RulesList
       classify_rules_list(dac, TestExList, RulesList, ClassIndex, ClassTuples, ClassTuples, ClassDist, Settings, Res),
       statistics(runtime, [_,EnsEvalTime]),
       length(Res, ResL),
       %write('Before evaluate'),nl,
       evaluate(ClassDist, ClassTuples, Res, ResL, PartEval).%,
       %write('After evaluate'),nl.
classify_rules(DataName, Method, TestExList, ClassIndex, ClassTuples, Keys, NoRules, NoConds, _Rules,
	       [result(DataName, MethodName, Size, Keys, ClassTuples, NoRules, NoConds, PartEval, EnsEvalTime)]):-
       ((Method = ensemble_model(Size, Settings, sac),
	 MethodName = ensemble_sac_bb_rules
	)
       ;
       (Method = single_model(Settings, sac),
	 MethodName = single_sac_bb_rules,
	 Size = 1
        )
       ),
       member(bb_rules, Settings),!,
       %TODO:add setting to classify_test_set to cover wv(yes) case!!!
       %collect_conditions(Rules, ClassIndex, RulesList),
       statistics(runtime, _),
       classify_test_set(sac, TestExList, ClassIndex, ClassTuples, ClassTuples, ClassDist, Keys, Res),
       statistics(runtime, [_,EnsEvalTime]),
       %check_ties(Res, NoTies),
       %write('We found this amount of ties: ' ), write(NoTies),nl,
       length(Res, ResL),
       %trace,
       evaluate(ClassDist, ClassTuples, Res, ResL, PartEval).
classify_rules(DataName, Method, TestExList, ClassIndex, ClassTuples, Keys, NoRules, NoConds, Rules,
	       [result(DataName, MethodName, Size, Keys, ClassTuples1, NoRules, NoConds, PartEval, EnsEvalTime)]):-
       ((Method = ensemble_model(Size, Settings, sac),
	 MethodName = ensemble_sac_list_of_rules
	)
       ;
	(Method = single_model(Settings, sac),
	 MethodName = single_sac_list_of_rules,
	 Size = 1
        )
       ),!,
       %collect_conditions(Rules, ClassIndex, RulesList),
       copy_term(ClassTuples, ClassTuples1),
       copy_term(ClassTuples, ClassTuples2),
       copy_term(ClassTuples, ClassTuples3),
       copy_term(ClassTuples, ClassTuples4),
       statistics(runtime, _),
       classify_rules_list(sac, TestExList, Rules, ClassIndex, ClassTuples2, ClassTuples3, ClassDist, Settings, Res),
       statistics(runtime, [_,EnsEvalTime]),
       %check_ties(Res, NoTies),
       %write('We found this amount of ties: ' ), write(NoTies),nl,
       length(Res, ResL),
       evaluate(ClassDist, ClassTuples4, Res, ResL, PartEval).
classify_rules(DataName, Method, TestExList, ClassIndex, ClassTuples, Keys, NoRules, NoConds, Rules,
	       [result(DataName, MethodName, Size, Keys, ClassTuples1, NoRules, NoConds, PartEval, EnsEvalTime)]):-
       ((Method = ensemble_model(Size, Settings, concept_learning),
	 MethodName = ensemble_cl_list_of_rules
	)
       ;
	(Method = single_model(Settings, concept_learning),
	 MethodName = single_cl_list_of_rules,
	 Size = 1
        )
       ),!,
       %collect_conditions(Rules, ClassIndex, RulesList),
       copy_term(ClassTuples, ClassTuples1),
       copy_term(ClassTuples, ClassTuples2),
       copy_term(ClassTuples, ClassTuples3),
       copy_term(ClassTuples, ClassTuples4),
       statistics(runtime, _),
       classify_rules_list(cl, TestExList, Rules, ClassIndex, ClassTuples2, ClassTuples3, ClassDist, Settings, Res),
       statistics(runtime, [_,EnsEvalTime]),
       %check_ties(Res, NoTies),
       %write('We found this amount of ties: ' ), write(NoTies),nl,
       length(Res, ResL),
       evaluate(ClassDist, ClassTuples4, Res, ResL, PartEval).              

%%    
% classify rules list
%
%classify_rules_list(sac, [TestEx|TestExs], Rules, ClassIndex, ClassTuples, TempCTup, FCTup, [PartRes|Res])
classify_rules_list(_, [], _, _, _, ClassDist, ClassDist, _, []).
classify_rules_list(Method, [TestEx|TestExs], Rules, ClassIndex, ClassTuples, TempCTup, FCTup, Settings, [PartRes|Res]):-
    TestEx =..[F|Args],
    nth1(ClassIndex, Args, Class, Rest),
    nth1(ClassIndex, NewArgs, _, Rest),
    %if we have not seen it before add it to tuple, fix this for rest...
    (select(No-Class, TempCTup, TCTup) ->
      NewNo is No + 1,
      append([NewNo-Class], TCTup, NewTempCTup)
    ;
     append([1-Class], TempCTup, NewTempCTup)
    ),
    NTestEx =..[F|NewArgs],
    %write(TestEx),nl,
    get_rules_votes1(Rules, Method, Settings, NTestEx, ClassIndex, Class-ClassTuples, PartRes, _RulesFiredList),!,
    %write('Predicted Class: '), write(PartRes),nl,
    %write('These rules fired: '),write(RulesFiredList), nl, 
    %nl,
    %(TestEx = row(i,0.225,0.17,0.07,0.0565,0.024,0.013,0.016,4) -> trace;true),
    classify_rules_list(Method, TestExs, Rules, ClassIndex, ClassTuples, NewTempCTup, FCTup, Settings, Res).
classify_rules_list(Method, [TestEx|TestExs], Rules, ClassIndex, ClassTuples, TempCTup, FCTup, Settings, [PartRes|Res]):-
	write('Failure in classify_rules_list'),nl,
	trace,
	classify_rules_list(Method, [TestEx|TestExs], Rules, ClassIndex, ClassTuples, TempCTup, FCTup, Settings, [PartRes|Res]).

get_rules_votes1([], dac, _, _, _, C-CT, C-CT,[]).
get_rules_votes1([tree(TreeRules)|RestRules], dac, Settings, NTestEx, ClassIndex, Class-FClassTuples, Class-RSRes, Dummy):-
    (get_rules_votes(TreeRules, dac, Settings, NTestEx, ClassIndex, Class-FClassTuples, _-PartRes) ->       %One rule per tree! 
     NPartRes = PartRes
    ;
     copy_term(FClassTuples, CClassTuples),
     NPartRes = CClassTuples
    ),
    get_rules_votes1(RestRules, dac, Settings, NTestEx, ClassIndex, Class-FClassTuples, _-RestRes, Dummy),
    joint_dist(RestRes, NPartRes, Res),
    keysort(Res, SRes),
    reverse(SRes, RSRes).

get_rules_votes1([], sac, _, _, _, C-CT, C-CT, []).
get_rules_votes1([rule_set(Rules)|Rest], sac, Settings, NTestEx, ClassIndex, Class-ClassTuples, Class-RSRes, FRulesFiredList):-
    (get_rules_votes(Rules, sac, Settings, NTestEx, ClassIndex, Class-ClassTuples, _-PartRes, RuleIDL) ->
     NPartRes = PartRes
    ;
     %failed to classify it...use no special value for it.
     copy_term(ClassTuples, CClassTuples),
     RuleIDL = [],
     NPartRes = CClassTuples
    ),
    get_rules_votes1(Rest, sac, Settings, NTestEx, ClassIndex, Class-ClassTuples, _-RestRes, RestRulesFired),
    append(RuleIDL, RestRulesFired, FRulesFiredList),
    joint_dist(RestRes, NPartRes, Res),
    keysort(Res, SRes),
    reverse(SRes, RSRes).

is_zero([]).
is_zero([0-_|R]):-
	is_zero(R).

%Not fixed for id
/*
get_rules_votes([Rule|Rules], dac, Settings, NTestEx, ClassIndex, TClass-ClassTuples, CTuples):-
    copy_term(ClassTuples, CClassTuples),	
    get_rule_vote(Rule, Settings, NTestEx, CClassTuples, ClassDist),
    (ClassDist = CClassTuples ->
      get_rules_votes(Rules, dac, Settings, NTestEx, ClassIndex, TClass-ClassTuples, CTuples)             %continue trying rules
    ;
      %TODO: fix wv(all) och wv(majority)
      keysort(ClassDist,Classes),
      reverse(Classes,[_-Class|_]),
      select(No-Class, ClassTuples, Rest),
      NewNo is No + 1,
      append([NewNo-Class], Rest, NewClassTuples),
      keysort(NewClassTuples, SNewClassTuples),
      reverse(SNewClassTuples, RSNewClassTuples),
      CTuples = TClass-RSNewClassTuples
    ).
*/
get_rules_votes([Rule|Rules], dac, Settings, NTestEx, ClassIndex, TClass-ClassTuples, CTuples):-
    %trace,	
    copy_term(ClassTuples, CClassTuples),	
    get_rule_vote(Rule, Settings, NTestEx, CClassTuples, ClassDist),
    (ClassDist = CClassTuples ->
      get_rules_votes(Rules, dac, Settings, NTestEx, ClassIndex, TClass-ClassTuples, CTuples)    %continue trying rules
    ;
      (member(wv(all), Settings)->
      joint_dist(ClassDist, ClassTuples, EvalClassT),
      keysort(EvalClassT, NewEvalClassT),
      reverse(NewEvalClassT, RSNewClassTuples)
    ;
     (member(wv(majority), Settings)->
      keysort(ClassDist,Classes),
      reverse(Classes,[MajNo-Class|_]),
      select(TDistNo-Class, ClassTuples, Rest),
      NewNo is MajNo + TDistNo,
      append([NewNo-Class], Rest, NewClassTuple),
      keysort(NewClassTuple, SNewClassTuple),
      reverse(SNewClassTuple, RSNewClassTuples)
     ;
      keysort(ClassDist,Classes),
      reverse(Classes,[_-Class|_]),
      select(No-Class, ClassTuples, Rest),
      NewNo is No + 1,
      append([NewNo-Class], Rest, NewClassTuple),
      keysort(NewClassTuple, SNewClassTuple),
      reverse(SNewClassTuple, RSNewClassTuples)
     )
    ),!,
      %keysort(ClassDist,Classes),
      %reverse(Classes,[_-Class|_]),
      %select(No-Class, ClassTuples, Rest),
      %NewNo is No + 1,
      %append([NewNo-Class], Rest, NewClassTuples),!,
      %keysort(NewClassTuples, SNewClassTuples),
      %reverse(SNewClassTuples, RSNewClassTuples),
      %get_rules_votes(Rules, sac, Settings, NTestEx, ClassIndex, TClass-RSNewClassTuples, CTuples, RuleIds),
      %append([Id],RuleIds,FRulesIds)
      %TODO: fix wv(all) och wv(majority)
      %keysort(ClassDist,Classes),
      %reverse(Classes,[_-Class|_]),
      %select(No-Class, ClassTuples, Rest),
      %NewNo is No + 1,
      %append([NewNo-Class], Rest, NewClassTuples),
      %keysort(NewClassTuples, SNewClassTuples),
      %reverse(SNewClassTuples, RSNewClassTuples),
      CTuples = TClass-RSNewClassTuples
    ).

get_rules_votes([], sac, _, _NTestEx, _ClassIndex, Class-ClassTuples, Class-RSClassT, []):-
    sort(ClassTuples, SClassT),
    reverse(SClassT, RSClassT).
get_rules_votes([rule(Id, Conds, RDist)|Rules], sac, Settings, NTestEx, ClassIndex, TClass-ClassTuples, CTuples, FRulesIds):-          %try all rules!
    copy_term(ClassTuples, CClassTuples),	
    get_rule_vote(rule(Id, Conds, RDist), Settings, NTestEx, CClassTuples, ClassDist),
    (ClassDist = CClassTuples ->
      get_rules_votes(Rules, sac, Settings, NTestEx, ClassIndex, TClass-ClassTuples, CTuples, FRulesIds)
    ;
      (member(wv(all), Settings)->
      joint_dist(ClassDist, ClassTuples, EvalClassT),
      keysort(EvalClassT, NewEvalClassT),
      reverse(NewEvalClassT, RSNewClassTuples)
    ;
     (member(wv(majority), Settings)->
      keysort(ClassDist,Classes),
      reverse(Classes,[MajNo-Class|_]),
      select(TDistNo-Class, ClassTuples, Rest),
      NewNo is MajNo + TDistNo,
      append([NewNo-Class], Rest, NewClassTuple),
      keysort(NewClassTuple, SNewClassTuple),
      reverse(SNewClassTuple, RSNewClassTuples)
     ;
      keysort(ClassDist,Classes),
      reverse(Classes,[_-Class|_]),
      select(No-Class, ClassTuples, Rest),
      NewNo is No + 1,
      append([NewNo-Class], Rest, NewClassTuple),
      keysort(NewClassTuple, SNewClassTuple),
      reverse(SNewClassTuple, RSNewClassTuples)
     )
    ),!,
      %keysort(ClassDist,Classes),
      %reverse(Classes,[_-Class|_]),
      %select(No-Class, ClassTuples, Rest),
      %NewNo is No + 1,
      %append([NewNo-Class], Rest, NewClassTuples),!,
      %keysort(NewClassTuples, SNewClassTuples),
      %reverse(SNewClassTuples, RSNewClassTuples),
      get_rules_votes(Rules, sac, Settings, NTestEx, ClassIndex, TClass-RSNewClassTuples, CTuples, RuleIds),
      append([Id],RuleIds,FRulesIds)
    ).

get_rule_vote(rule(_, [], _), _, _, TClassTuple, TClassTuple).                       %Ignore default rule
get_rule_vote(rule(Id, Conds, Dist), Setting, Ex, TClassTuple, NewTClassTuple):- %SAC	
    update_dist(Conds, Id, Setting, Ex, Dist, TClassTuple, NewTClassTuple).
get_rule_vote(rule(Conds, Dist), Setting, Ex, TClassTuple, NewTClassTuple):-    %DAC , no rule id...fix this
    %write(Conds),nl,
    %trace,
    update_dist(Conds, _Id, Setting, Ex, Dist, TClassTuple, NewTClassTuple).

update_dist([], _Id, Setting, _Ex, Dist, TClassTuple, RSNewClassTuple):-
    %write('Rule with id: '),write(Id), write(' fired!'),nl,
    (member(wv(all), Setting)->
      joint_dist(Dist, TClassTuple, EvalClassT),
      keysort(EvalClassT, NewEvalClassT),
      reverse(NewEvalClassT, RSNewClassTuple)
    ;
     (member(wv(majority), Setting)->
      keysort(Dist,Classes),
      reverse(Classes,[MajNo-Class|_]),
      select(TDistNo-Class, TClassTuple, Rest),
      NewNo is MajNo + TDistNo,
      append([NewNo-Class], Rest, NewClassTuple),
      keysort(NewClassTuple, SNewClassTuple),
      reverse(SNewClassTuple, RSNewClassTuple)
     ;
      keysort(Dist,Classes),
      reverse(Classes,[_-Class|_]),
      select(No-Class, TClassTuple, Rest),
      NewNo is No + 1,
      append([NewNo-Class], Rest, NewClassTuple),
      keysort(NewClassTuple, SNewClassTuple),
      reverse(SNewClassTuple, RSNewClassTuple)
     )
    ).
    
update_dist([index(I)=CVal|Conds], Id, Settings, Ex, Dist, TClassTuple, NewTClassTuple):-  
    arg(I, Ex, ExVal),
    ExVal \== '?',	
    (ExVal = CVal ->
      update_dist(Conds, Id, Settings, Ex, Dist, TClassTuple, NewTClassTuple)        %okay, so continue
    ;
      TClassTuple = NewTClassTuple                                    %Failed so stop asap
    ).
update_dist([index(I)<CVal|Conds], Id, Settings, Ex, Dist, TClassTuple, NewTClassTuple):-  
    arg(I, Ex, ExVal),
    ExVal \== '?',	
    (ExVal < CVal ->
      update_dist(Conds, Id, Settings, Ex, Dist, TClassTuple, NewTClassTuple)        %okay, so continue
    ;
      TClassTuple = NewTClassTuple                                    %Failed so stop asap
    ).
update_dist([index(I)>=CVal|Conds], Id, Settings, Ex, Dist, TClassTuple, NewTClassTuple):-  
    arg(I, Ex, ExVal),
    ExVal \== '?',	
    (ExVal >= CVal ->
      update_dist(Conds, Id, Settings, Ex, Dist, TClassTuple, NewTClassTuple)        %okay, so continue
    ;
      TClassTuple = NewTClassTuple                                     %Failed so stop asap
    ).

evaluate([], _, [], _, []).
evaluate([No-Class|CTuples], ClassTuple, Res, ResL, [Class-[No, TrueP, FalseP, FalseN, FalseTup]|Results]):-
    evaluate_class(Res, Class, RestL, 0, 0, 0, ClassTuple, TrueP, FalseP, FalseN, FalseTup),
    %PercentTrueP is (TrueP / No) * 100,   %ResL
    %PercentFalseP is (FalseP / ResL) * 100, %ResL
    %ClassCov is (No / ResL) * 100,
    %nl,
    %write('Class: '), write(Class),nl,
    %write('Covers '), write(ClassCov), write(' of the example space'),nl,
    %write('Accuracy: '), write(PercentTrueP),nl,
    %write('% of False Classifications(wrt. to all classes): '), write(PercentFalseP),nl,
    %write_false_neg1(FalseTup, ResL, FNegL),!,
    evaluate(CTuples, ClassTuple, RestL, ResL, Results).

evaluate_class([], _Class, [], TrueP, FalseP, FalseN, False, TrueP, FalseP, FalseN, False).
evaluate_class([Class-[_-Class|_]|Rest], Class, RestL, TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    NewTTrueP is TTrueP + 1,!,
    evaluate_class(Rest, Class, RestL, NewTTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False).
evaluate_class([Class2-[No-Class|R]|Rest], Class, [Class2-[No-Class|R]|RestL], TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    Class \== Class2,!,
    NewTFalseP is TFalseP + 1,
    update_false(TFalse, Class, _NewTFalse),
    evaluate_class(Rest, Class, RestL, TTrueP, NewTFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False).
evaluate_class([Class-[_-Class2|_]|Rest], Class, RestL, TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    Class \== Class2,!,
    NewTFalseN is TFalseN + 1,
    update_false(TFalse, Class2, NewTFalse),
    evaluate_class(Rest, Class, RestL, TTrueP, TFalseP, NewTFalseN, NewTFalse, TrueP, FalseP, FalseN, False).
evaluate_class([Class2-[No-Class3|R]|Rest], Class, [Class2-[No-Class3|R]|RestL], TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    evaluate_class(Rest, Class, RestL, TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False).

update_false(TFalseN, Class, NewTFalseN):-
    select(No-Class, TFalseN, RestTFalseN),
    NewNo is No + 1,!,
    append([NewNo-Class], RestTFalseN, NewTFalseN).

check_ties(Res, NoTies):-
    check_ties(Res, 0, NoTies).
check_ties([], NoTies, NoTies).
check_ties([_-[No-_,No-_|_]|Rest], TempNoT, NoTies):-
    NewTempNoT is TempNoT + 1,!,
    check_ties(Rest, NewTempNoT, NoTies).
check_ties([_|Rest], TempNoT, NoTies):-
    check_ties(Rest, TempNoT, NoTies).