%%%%
% Read from File
% Module that reads in data and does som pre-processing
%
% Author: Tony Lindgren 
%
% version 0.0.1
%


:- module(read_from_file,[read_from_file/2, interpret_info/4, get_csv_examples/7, pre_process_data/4]).

:- use_module([library(csv), library(lists), library(random), library(codesio), util]).

% Read clauses from file
read_from_file(FileName,FileInformation):-
    exists_file(FileName),!,
    see(FileName),
    read(X),
    read_loop(X,FileInformation),
    seen.
read_from_file(_FileName,_FileInformation):-
    write('No such file exist'),nl.

read_loop(end_of_file,[]).
read_loop(X,[X|R]):-
    read(NewX),
    read_loop(NewX,R).

interpret_info(InfoList, Head, BackgroundKnowledge, Examples):-
   member(background_head(Name/Arity), InfoList),
   functor(Head, Name, Arity),
   member(background_file(BackgroundKnowledge), InfoList),
   member(example_file(Examples), InfoList).

pre_process_data(N, Params, ExList, ExLists):- 
    length(ExList, NoEx),
    ExPerFold is round(NoEx / N),
    (member(validation_size(VSize),Params) ->
      NoValPerFold is round(ExPerFold * VSize)
     ;
     NoValPerFold is 0
    ),
    NewExPerFold is ExPerFold - NoValPerFold,!,
    rnd_fold_gen(N, ExList, NewExPerFold, _NoValPerFold, NoEx, ExLists). %, ValExList).

rnd_fold_gen(0, [], _ExPerFold, _NoValPerFold, _NoEx, []). %, []).
rnd_fold_gen(0, _ExLeft, _, _, _, []). %, []). % throw example away!, fix better solution later
rnd_fold_gen(N, Examples, ExPerFold, NoValsPerFold, NoEx, [Fold|Folds]):- %, [FoldVal|FoldVals]):-
    %gen_one_fold(NoValsPerFold, NoEx, Examples, NewNoEx, NewExamples, FoldVal),
    gen_one_fold(ExPerFold, NoEx, Examples, NewNoEx, NewExamples, Fold),
    NewN is N - 1,!,
    rnd_fold_gen(NewN, NewExamples, ExPerFold, NoValsPerFold, NewNoEx, Folds). %, FoldVals).

gen_one_fold(0, NoEx, Examples, NoEx, Examples, []).
gen_one_fold(_, 0, Examples, 0, Examples, []).
gen_one_fold(ExPerFold, NoEx, Examples, RestNoEx, RestExamples, [Ex|Exs]):-
    %Pick training examples per fold randomly
    random(0, NoEx, Rand),
    nth0(Rand, Examples, Ex, NewExamples),
    NewExPerFold is ExPerFold - 1,
    NewNoEx is NoEx - 1,!,
    gen_one_fold(NewExPerFold, NewNoEx, NewExamples, RestNoEx, RestExamples, Exs).

get_csv_examples(CsvParameters, Parameters, NewParameters, 1, Head, [ExTrainList,ExTestList], FeaturesDTL):-   
    member(testfile(TestFileName), CsvParameters),
    member(trainfile(TrainFileName), CsvParameters),!,
    (member(mat(ClassI, Rows, Cols, Type), CsvParameters) ->
     get_mat_data(TestFileName, TrainFileName, ClassI, Rows, Cols, Type, Parameters, NewParameters, Head, ExTrainList, ExTestList, FeaturesDTL)
    ;
     append([strip(true)], CsvParameters, NCSVParameters),
     csv_read_file(TrainFileName, [FirstTrainRow|ExTrainList], NCSVParameters),
     csv_read_file(TestFileName, [_|ExTestList], NCSVParameters),	                        
     select(fold_x_v(_), Parameters, RestParameters), %ignore whate ever we have here and set it to 1
     append([folds_x_v(1)], RestParameters, NewParameters),
     get_csv_head_and_features(FirstTrainRow, Head, PrevFeaturesDTL),
     handle_categoric_d(PrevFeaturesDTL, 1, ExTrainList, FeaturesDTL)
    ).    
get_csv_examples(CsvParameters, Parameters, Parameters, Folds, Head, ExLists, FeaturesDTL):-    %use only one file 
    member(filename(FileName), CsvParameters),!,
    append([strip(true)], CsvParameters, NCSVParameters),
    csv_read_file(FileName, [FirstRow|ExList], NCSVParameters), 
    random_permutation(ExList, RndExL),	                        
    member(fold_x_v(Folds), Parameters),
    pre_process_data(Folds, Parameters, RndExL, ExLists),
    get_csv_head_and_features(FirstRow, Head, PrevFeaturesDTL),
    handle_categoric_d(PrevFeaturesDTL, 1, ExList, FeaturesDTL).

get_csv_head_and_features(Head, NewHead, RFeatureDTList):-
  functor(Head, Functor, Args),
  functor(NewHead, Functor, Args), 
  csv_feature_data_type(Args, Head, NewHead, FeatureDTList),
  reverse(FeatureDTList, RFeatureDTList).

csv_feature_data_type(0, _, _, []).
csv_feature_data_type(Arg, Head, NewHead, [Arg-PartBody|RestBody]):-
    arg(Arg, Head, TypeAndName),
    atom_codes(TypeAndName, TypeAndNameC),	
    split_to_parts(TypeAndNameC, [], Type, _Name), %use Name later on if needed..	
    arg(Arg, NewHead, Var),
    PartBody =.. [Type,Var],
    NewArg is Arg - 1,!,
    csv_feature_data_type(NewArg, Head, NewHead, RestBody).

split_to_parts([45|Rest], Prev, Type, Name):-
    reverse(Prev, RPrev),	
    atom_codes(Type, RPrev),
    atom_codes(Name, Rest).
split_to_parts([CharCode|Rest], Prev, Type, Name):- 
    split_to_parts(Rest, [CharCode|Prev], Type, Name).

handle_categoric_d([], _, _, []).
handle_categoric_d([I-categoric(Att)|RestFDTL], Arg, ExList, [I-categoric(Att, Categories)|FDTLs]):-
    get_categories(ExList, Arg, [], Categories),
    NewArg is Arg + 1,!,	
    handle_categoric_d(RestFDTL, NewArg, ExList, FDTLs).
handle_categoric_d([Att|RestFDTL], Arg, ExList, [Att|FDTLs]):-
    NewArg is Arg + 1,!,
    handle_categoric_d(RestFDTL, NewArg, ExList, FDTLs).

get_categories([], _, TempCat, TempCat).
get_categories([Ex|Exs], Arg, TempCat, Categories):-
    arg(Arg, Ex, CatVal),!,
    (member(CatVal, TempCat) ->
     get_categories(Exs, Arg, TempCat, Categories)
    ;
     get_categories(Exs, Arg, [CatVal|TempCat], Categories)
    ).

get_mat_data(TestFileName, TrainFileName, ClassI, Rows, Cols, Type, Parameters, NewParameters, mat(ClassI,Rows,Cols,Type), ExTrainMatList, ExTestMatList, mat(ClassI,Rows,Cols,Type)):-
    NCSVParameters = [strip(true)],
    csv_read_file(TrainFileName, ExTrainList, NCSVParameters),	                        
    csv_read_file(TestFileName, ExTestList, NCSVParameters),
    select(fold_x_v(_), Parameters, RestParameters), %ignore whate ever we have here and set it to 1
    append([folds_x_v(1)], RestParameters, NewParameters),
    convert_ex_to_mat(ExTrainList, ClassI, Rows, Cols, Type, ExTrainMatList),
    convert_ex_to_mat(ExTestList, ClassI, Rows, Cols, Type, ExTestMatList).

convert_ex_to_mat([], _, _, _, _, []).
convert_ex_to_mat([Ex|Exs], ClassI, Rows, Cols, _, [mat_ex(MatEx, ClassVal)|ExTrainMatList]):-
    mat(MatEx, Rows, Cols),
    Ex =.. [_|Args],
    nth1(ClassI, Args, ClassVal, RestArgs), 
    fill_mat(RestArgs, 1, 1, Rows, Cols, MatEx),
    convert_ex_to_mat(Exs, ClassI, Rows, Cols, _, ExTrainMatList).

fill_mat([], _, _, _, _, _).
fill_mat([Arg|Args], CRow, EndCols, EndRows, EndCols, MatEx):-
    mat(MatEx, CRow, EndCols, Arg),
    NewCRow is CRow + 1,
    fill_mat(Args, NewCRow, 1, EndRows, EndCols, MatEx).	
fill_mat([Arg|Args], CRow, CCol, EndRows, EndCols, MatEx):-
    mat(MatEx, CRow, CCol, Arg),
    NewCCol is CCol + 1, 	
    fill_mat(Args, CRow, NewCCol, EndRows, EndCols, MatEx).