%%
% Ensamble induction method using DAC
%
% Author: Tony Lindgren
%

:- module(tree,[tree/10, create_class_tuples/2]).
:- use_module([library(lists), library(random), util]).



tree(TrainExList, no_bk, Parameters, Head, _, ClassIndex, FeaturesDTL, ClassTupleC, Tree, CHead):-
    %write('Starting to build tree...'),nl,
    %trace,
    member(classification, Parameters),
    nth1(ClassIndex, FeaturesDTL, ClassIndex-class(_)),
    find_categories(TrainExList, ClassIndex, Classes, ClassList),
    class_tuple_c(Classes, ClassTupleC),
    Tree = dac_node(Head, FeaturesDTL, ClassList, _Children),!,
    build_tree(TrainExList, no_bk, Parameters, ClassIndex, ClassTupleC, first, Tree),
    %write('Finished building tree...'),nl,
    %write('Tree: '),write(Tree),nl,
    %trace,
    keysort(ClassList, SortedTuples),
    reverse(SortedTuples, [_-DefaultClass|_]),
    copy_term(Head,CHead),
    CHead =.. [_|Args],
    nth1(ClassIndex, Args, DefaultClass).

tree(TrainExList, BKList, Parameters, Head, BBKey, ClassIndex, no_f, ClassT, Tree, CHead):-
   %write('Starting to build tree...'),nl,
    %trace, 	
    assert_bb(BKList,BBKey), %assert background knowledge, used when building the tree
    %trace,
    feature_data_type(Head, BBKey, FeatureDTList),
    member(classification, Parameters),
    count_classes(TrainExList, BBKey, FeatureDTList, ClassIndex, ClassTupleC, ClassList),!,
    c_index_features(FeatureDTList, 1, IFeatureDTList), 
    Tree = dac_node(Head, IFeatureDTList, ClassList, _Children),
    build_tree(TrainExList, BBKey, Parameters, ClassIndex, ClassTupleC, first, Tree),
    %write('finished building tree...cleaning up..'),nl,
    create_class_tuples(BBKey, ClassT),
    retract_bb([Head], BBKey),
    keysort(ClassList, SortedTuples),
    reverse(SortedTuples, [_-DefaultClass|_]),
    copy_term(Head,CHead),
    CHead =.. [_|Args],
    nth1(ClassIndex, Args, DefaultClass).

c_index_features([], _, []). 
c_index_features([F|FDTL], No, [No-F|IFDTList]):-
    NewNo is No + 1,
    !,c_index_features(FDTL, NewNo, IFDTList).

%dupliate here and rule_set move to util..how to handle first time see sac implementation, now probably wrong...
find_categories(TrainEx, ClassIndex, Classes, ClassTupleC):-
    find_categories(TrainEx, ClassIndex, [], Classes, [], ClassTupleC).
find_categories([], _, Classes, Classes, ClassTupleC, ClassTupleC).
find_categories([Ex|Exs], ClassIndex, TClasses, Classes, TClassTupleC, ClassTupleC):-
    arg(ClassIndex, Ex, Class),
    (member(Class, TClasses) ->
      select(No-Class, TClassTupleC, RestTClassTupleC),
      NewNo is No + 1,!,
      find_categories(Exs, ClassIndex, TClasses, Classes, [NewNo-Class|RestTClassTupleC], ClassTupleC)
     ;
      !,find_categories(Exs, ClassIndex, [Class|TClasses], Classes, [1-Class|TClassTupleC], ClassTupleC)
    ).

build_tree([], _, _, _, _, _, dac_node(_, _, _,no_examples_left)):-!.
build_tree(TrainExList,_BBKey, Parameters, _, _, _, dac_node(_,_,_,min_cov)):-
    %trace,	
    length(TrainExList, NoEx),
    member(min_cov(MinCov),Parameters),
    %write('Min cov Leaf'),
    MinCov >= NoEx,!.
build_tree(TrainExList, BBKey, Parameters, ClassIndex, ClassTupleL, PrevBiggestCP, DACNode):-	
    length(TrainExList, NoEx),
    DACNode = dac_node(_,_,ClassList,_),
    okay_to_stop(ClassList, NoEx, Parameters, PrevBiggestCP, NewBiggestCP, Status),
    (Status = cont ->    
      !,c_build_tree(TrainExList, BBKey, Parameters, ClassIndex, ClassTupleL, NewBiggestCP, DACNode)              
    ;
      %write('clean engough'),nl,
      DACNode = dac_node(_,_,ClassList,clean_enough)
    ).	
c_build_tree(TrainExList, BBKey, Parameters, ClassIndex, ClassTupleL, CP, dac_node(_Head,FeatureDTList,ClassList,ChildrenNodes)):-	
    %write('Splitting'),nl,
    reset_c_t(ClassTupleL, ResetCTL),
    %select_feature(FeatureDTList, TrainExList, ResetCTL, ClassList, ClassIndex, 1, FeatureRanking),!,
    %select_values(FeatureRanking, BBKey-ResetCTL, Parameters, FeatureDTList, ClassIndex, TrainExList, ChildrenNodes).
    sel_feature(FeatureDTList, TrainExList, ResetCTL, ClassList, ClassIndex, Parameters, first-_-_, F-List-_),!,
    get_att(F, NewF),
    apply_split(List, BBKey-ResetCTL, Parameters, FeatureDTList, ClassIndex, TrainExList, NewF, CP, ChildrenNodes).

reset_c_t([], []).
reset_c_t([_-Class|Rest], [0-Class|ResetR]):-
	reset_c_t(Rest, ResetR).

okay_to_stop(ClassList, NoEx, Parameters, PrevBiggestCP, NewBiggestCP, Status):-
    check_class_margin(ClassList, NoEx, MarginL),
    sort(MarginL, SortedMargin),
    check_margins(SortedMargin, PrevBiggestCP, NewBiggestCP, Parameters, Status).
   
check_margins(SortedCP, first, CP1, Parameters, Status):-
    member(min_margin(MMargin), Parameters),
    reverse(SortedCP, [CP1-_,CP2-_|_]),
    Margin is abs(CP1 - CP2),
    %(var(MMargin)->trace;true),
    (Margin > MMargin ->
     %write('Stopped because clean_enough'),nl,
     !,Status = stop
     ;
       Status = cont
    ).	
check_margins(SortedCP, PrevBiggestCP, CP1, Parameters, Status):-
    member(min_margin(MMargin), Parameters),
    reverse(SortedCP, [CP1-_,CP2-_|_]),
    Margin is abs(CP1 - CP2),
    %(var(MMargin)->trace;true),
    (Margin > MMargin ->
     %write('Stopped because clean_enough'),nl,
     !,Status = stop
    ;
     (member(rel_margin(RelMargin), Parameters) ->
       %reverse(SortedCP, [CP1-_|_]),
       NewBiggestCP is CP1 + RelMargin,   %0 = must be stictly better, +Value = we can be worse tha previously, -Value = we must be so much better
       (PrevBiggestCP > NewBiggestCP ->
        !,%write('Tree:Stopped because rel_margin'),write(PrevBiggestCP),write('  '),write(NewBiggestCP),write('  '),write(RelMargin),nl,
        Status = stop
       ;
        !,Status = cont
       )
     ;
       Status = cont
     )
    ).

check_class_margin([],_,[]).
check_class_margin([Count-_|Counts],NoTot,[Percent-_|Ps]):-
	Percent is Count / NoTot,
	check_class_margin(Counts,NoTot,Ps).

%%	This baby should call build tree, be sure to add all children to
				%	the tree...fix a stopping criterion at build tree as well...
/*
select_values(FeatureRanking, BBKey, Param, FeatureDTList,  ClassIndex, TrainExList, ChildrenNodes):-
        (member(rnd_feature(yes), Param) ->
           length(FeatureRanking, FLength),
           repeat,             
             random(0, FLength, Rnd),
             FeatureSel is  Rnd + 1,
             FeatureSel \== ClassIndex,   
           nth1(FeatureSel,FeatureRanking,F1-List1-_),
           F = F1,
           List = List1
	 ;  
	   find_highest_gain(FeatureRanking, F2-List2-_),
           F = F2,
           List = List2
        ),!,
        get_att(F, NewF),
	apply_split(List, BBKey, Param, FeatureDTList, ClassIndex, TrainExList, NewF, ChildrenNodes).
*/
get_att(numeric(F), F).
get_att(categoric(F,_), F).

%Only works for numerical, right now
find_highest_gain(GainList,Result):-
	first_with_value(GainList, First, Rest),
	find_highest_gain(Rest,First,Result).
find_highest_gain([],Result,Result).
find_highest_gain([_-_-first-_|Rest],Temp,Result):- 
	%ignore dummy values
	find_highest_gain(Rest,Temp,Result).
find_highest_gain([Feature-Split-Gain|Rest],_-_-TGain,Result):-
	TGain < Gain,!,
	find_highest_gain(Rest,Feature-Split-Gain,Result).
find_highest_gain([_|Rest],Temp,Result):-
	find_highest_gain(Rest,Temp,Result).

first_with_value([_-_-first-_|RestGList], First, Rest):-
	!,first_with_value(RestGList, First, Rest).
first_with_value([First|Rest], First, Rest).

apply_split([], _, _, _, _, _, _, _, []).
apply_split([_-[[]-_]|_], no_bk-_, _, _, _, _, _, _, dac_node(_,_,_,cannot_split)).% :-
%	apply_split(RestCat, no_bk-ClassTL, Param, FeatureDTList, Index, TrainExList, F, CP, RestChildren).
apply_split([_-[_-LeftEx]|_], no_bk-_, _, _, _, TrainExList, _, _, dac_node(_,_,_,cannot_split)):-
	permutation(LeftEx, TrainExList).%,
 %       apply_split(RestCat, no_bk-ClassTL, Param, FeatureDTList, Index, TrainExList, F, CP, RestChildren).
apply_split([Cat-[InsDist-LeftEx]|RestCat], no_bk-ClassTL, Param, FeatureDTList, Index, TrainExList, F, CP, [LeftChildren|RestChildren]):-
	%write('Categoric case'),nl,trace,
	is_list(InsDist),
        LeftChildren = dac_node(F = Cat, FeatureDTList, InsDist, _LeftChildrenNodes),!,
        %write(LeftExamples),nl,
	build_tree(LeftEx, no_bk, Param, Index, InsDist, CP, LeftChildren),
        apply_split(RestCat, no_bk-ClassTL, Param, FeatureDTList, Index, TrainExList, F, CP, RestChildren).
	
apply_split([_, []], no_bk-_, _, _, _, _, _, _, dac_node(_,_,_,cannot_split)). %Cannot split any more...right
apply_split([[], _], no_bk-_, _, _, _, _, _, _, dac_node(_,_,_,cannot_split)). %Cannot split any more...left
apply_split([LeftSplitList, [Val2-(_,_)|_]], no_bk-ClassTL, Param, FeatureDTList, Index, TrainExList, F, CP,[LeftChildren,RightChildren]):-
        number(Val2),
	last(LeftSplitList, Val1-(_,_)),
	decimals(D),
	Diff is Val2 - Val1,
	Term is Diff / 2,
	Middle is round((Val1 + Term) * D) / D,!,    
        (Val1 =:= Val2 ->
            %trace,
	    %write('Single-'),
            Split is round(Val1 * D) / D
	 ;
 	    %write('Val2: '),write(Val2), write(' Val1: '),write(Val1),nl,
	    %write('Middle-'),
            Split = Middle
	    
        ),
	%write('LeftVal: '),write(Val1),nl,
	%write('RightVal: '),write(Val2),nl,
	%write('split: '),write(Split),nl,
	get_examples(LeftSplitList, TrainExList, LeftExamples, RightExamples),!,
        %length(LeftSplitList,LSPL), write(LSPL),nl,
        %length(LeftExamples,LEL), write(LEL),nl,
        %length(RightExamples,REL), write(REL),nl,
        %length(TrainExList,TEL), write(TEL),nl,
	%create_class_tuples(BBKey, ClassTL),
	%copy_term(ClassTL, CClassTL1),
	%copy_term(ClassTL, CClassTL2),
	%LeftSide
        do_count_classes(LeftExamples, Index, ClassTL, LClassTupleL),
	%write('Left ClassT: '),write(LClassTupleL),nl,
        LeftChildren = dac_node(F < Split, FeatureDTList, LClassTupleL, _LeftChildrenNodes),!,
        %write(LeftExamples),nl, 
	build_tree(LeftExamples, no_bk, Param, Index, LClassTupleL, CP,LeftChildren),
	%Right side
	do_count_classes(RightExamples, Index, ClassTL, RClassTupleL),
	%write('Right ClassT: '),write(RClassTupleL),nl,
        RightChildren = dac_node(F >= Split, FeatureDTList, RClassTupleL, _RightChildrenNodes),!,         
        %write(RightExamples),nl,
	build_tree(RightExamples, no_bk, Param, Index, RClassTupleL, CP, RightChildren).
apply_split([LeftSplitList, [Val2-(_,_)|_]], BBKey-_, Param, FeatureDTList, Index, TrainExList, F, CP, [LeftChildren,RightChildren]):-
	last(LeftSplitList, Val1-(_,_)),
        %trace,
	decimals(D),
	Diff is Val2 - Val1,
	Term is Diff / 2,
	Middle is round((Val1 + Term) * D) / D,!,    
        (Val1 =:= Val2 ->
            %trace,
	    %write('Single-'),
            Split is round(Val1 * D) / D
	 ;
 	    %write('Val2: '),write(Val2), write(' Val1: '),write(Val1),nl,
	    %write('Middle-'),
            Split = Middle
	    
        ),
	%write('LeftVal: '),write(Val1),nl,
	%write('RightVal: '),write(Val2),nl,
	%write('split: '),write(Split),nl,
	get_examples(LeftSplitList, TrainExList, LeftExamples, RightExamples),!,
        %length(LeftSplitList,LSPL), write(LSPL),nl,
        %length(LeftExamples,LEL), write(LEL),nl,
        %length(TrainExList,TEL), write(TEL),nl,
        %length(RL,REL), write(REL),nl,
        create_class_tuples(BBKey, ClassTL1),
	create_class_tuples(BBKey, ClassTL2),
        do_count_classes(LeftExamples, Index, ClassTL1, LClassTupleL),
        LeftChildren = dac_node(F < Split, FeatureDTList, LClassTupleL, _LeftChildrenNodes),!,
        %write(LeftExamples),nl, 
	build_tree(LeftExamples, BBKey, Param, Index, LClassTupleL, CP, LeftChildren),
	do_count_classes(RightExamples, Index, ClassTL2, RClassTupleL),
        RightChildren = dac_node(F >= Split, FeatureDTList, RClassTupleL, _RightChildrenNodes),!,         
        %write(RightExamples),nl,
	build_tree(RightExamples, BBKey, Param, Index, RClassTupleL, CP, RightChildren).

get_examples(SplitList, TrainExList, LeftList, RightList):-
	%trace,
        left_part(SplitList, TrainExList, LeftList),
        list_difference_eq(TrainExList, LeftList, RightList).

list_difference_eq([], _, []).
list_difference_eq([Ex|Train], LeftList, RightList):-
        member(Ex, LeftList),!,
        list_difference_eq(Train, LeftList, RightList).
list_difference_eq([Ex|Train], LeftList, [Ex|RightList]):-
        list_difference_eq(Train, LeftList, RightList).
         
left_part([], _, []).
left_part([_-(Key,_)|VKCs], TrainExList, [Ex|LeftEx]):-
	nth1(Key, TrainExList, Ex),
	left_part(VKCs, TrainExList, LeftEx).

%re-write keep only the best value, now we consul a lot of memory...see sac code
/*
select_feature([], _TrainExList, _, _ClassDist, _ClassIndex, _I, []).
select_feature([class(_)|FeatureDTList], TrainExList, ClassTupleL, ClassDist, ClassIndex, I, FSGs):- %Do not use class for splitt
	!,select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, I, FSGs).
select_feature([id(_)|FeatureDTList], TrainExList, ClassTupleL, ClassDist, ClassIndex, I, FSGs):- %Do not use id for splitt
	!,select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, I, FSGs).
select_feature([F|FeatureDTList], TrainExList, ClassTupleL, ClassDist, ClassIndex, I, [F-CatL-Gain|FSGs]):-
	F = categoric(_, Categories),!,
	%trace,
	%write('Innan split...'),nl,
	%write('ClassDist: '),write(ClassDist),nl,
	%write('TrainExList: '), write(TrainExList),
	%reset ClassT here?!
        cat_split(Categories, TrainExList, ClassDist, ClassTupleL, ClassIndex, I, DistL, CatL),
	i_gain(DistL, ClassDist, Gain),
	%atom_codes(Op,[61]), %61 is =
	%Split=[LS,RS],
	%last(LS, Val1-(_,_)),
	%RS=[Val2-(_,_)|_],
	%(Val1 = Val2 ->    %Do not split into impossible values!
	%   FGain = 0
	%;
	%   FGain = Gain
	%),	 
        %write('F-Split-Gain: '), write(F-Split-Gain),nl,
        %Split=[LS,RS],
        %length(LS,LenLs),
        %length(RS,LenRs), 
        %write('Left: '),write(LenLs),nl, 
        %write('Right: '),write(LenRs),nl, 
        NewI is I + 1,
	%write('I = '),write(I),nl,
	%(I = 3 -> trace; true),!,
	!,select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, NewI, FSGs).
select_feature([F|FeatureDTList], TrainExList, ClassTupleL, ClassDist, ClassIndex, I, [F-Split-FGain|FSGs]):-
	F = numeric(_),!, 
        split(TrainExList, ClassDist, _, ClassIndex, I, F, Split-Gain),
	Split=[LS,RS],
	(LS = [] ; RS = [] ->   % ignore empty sub_nodes...
	   FGain = 0
	;
	   last(LS, Val1-(_,_)),
	   RS=[Val2-(_,_)|_]
	),
	(Val1 = Val2 ->    %Do not split into impossible values!
	   FGain = 0
	;
	   FGain = Gain
	),	 
        %write('F-Split-Gain: '), write(F-Split-Gain),nl,
        %Split=[LS,RS],
        %length(LS,LenLs),
        %length(RS,LenRs), 
        %write('Left: '),write(LenLs),nl, 
        %write('Right: '),write(LenRs),nl, 
        NewI is I + 1,
	!,select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, NewI, FSGs).
*/
sel_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TFSG, FSG):-
        (member(rnd_feature(yes, SelParam), Param) ->         
	   length(FeatureDTList, FLength),
	   NoF is round(FLength * SelParam),
	   (NoF < 1 ->
	    NoFeatures = 1                                   %must be at least 1!
	   ;
	    NoF = NoFeatures
	   ),
	   rnd_f_sel(NoFeatures, FeatureDTList, FLength, ClassIndex, [], RndFDTList),
     	   select_feature(RndFDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TFSG, FSG)	   
	 ;  
	   select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TFSG, FSG)
        ).

%do not handle id:s for now - now using 
rnd_f_sel(0, _, _, _, _, []).
rnd_f_sel(NoFeatures, FeatureDTList, FLength, ClassIndex, HList, [RndSelection|RndFDTList]):-           
       random(0, FLength, Rnd),	
       FeatureSel is Rnd + 1,
       \+ member(FeatureSel, HList), 
       FeatureSel \== ClassIndex,!,
       nth1(FeatureSel,FeatureDTList,RndSelection),
       NewNoFeatures is NoFeatures - 1,
       rnd_f_sel(NewNoFeatures, FeatureDTList, FLength, ClassIndex, [FeatureSel|HList], RndFDTList).
rnd_f_sel(NoFeatures, FeatureDTList, FLength, ClassIndex, HList, RndSelection):-           %try again...
       rnd_f_sel(NoFeatures, FeatureDTList, FLength, ClassIndex, HList, RndSelection).

select_feature([], _, _, _, _, _, BestSplitForThisFeature, BestSplitForThisFeature).
select_feature([_-class(_)|FeatureDTList], TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TFSG, FSG):-
        !,select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TFSG, FSG).
select_feature([_-id(_)|FeatureDTList], TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TFSG, FSG):- %Do not use id for splitt
	!,select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TFSG, FSG).
select_feature([I-F|FeatureDTList], TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, first-_-_, FSG):-
	%add handeling of random features...in params...later
	F = categoric(_, Categories),!,
	%write('Innan categoric split...'),nl,
	%write('ClassDist: '),write(ClassDist),nl,
	%write('TrainExList: '), write(TrainExList),
	cat_split(Categories, TrainExList, ClassDist, ClassTupleL, ClassIndex, I, DistL, CatL),!,
	(member(cat_split(all), Param) ->
	  i_gain(DistL, ClassDist, NGain),
	  NCatL = CatL
	;
	  %trace,
	  best_binary_i_gain(DistL, ClassDist, NGain, CatL, NCatL)
	),
	(NGain = first ->
	 Gain = 0
	;
	 Gain = NGain
	),
	%atom_codes(Op,[61]), %61 is =
	%Split=[LS,RS],
	%last(LS, Val1-(_,_)),
	%RS=[Val2-(_,_)|_],
	%(Val1 = Val2 ->    %Do not split into impossible values!
	%   FGain = 0
	%;
	%   FGain = Gain
				%),
	(Gain = first ->
	 NGain = 0
	;
	 NGain = Gain
	),
        %write('Gain: '), write(Gain),nl,

        %Split=[LS,RS],
        %length(LS,LenLs),
        %length(RS,LenRs), 
        %write('Left: '),write(LenLs),nl, 
        %write('Right: '),write(LenRs),nl, 
	%write('I = '),write(I),nl,
	%(I = 3 -> trace; true),!,
	select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, F-NCatL-Gain, FSG).
select_feature([I-F|FeatureDTList], TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TF-TSplit-TGain, FSG):-
	%add handeling of random features...in params...later
	F = categoric(_, Categories),!,
	%write('Innan categoric split...'),nl,
	%write('ClassDist: '),write(ClassDist),nl,
	%write('TrainExList: '), write(TrainExList),
        cat_split(Categories, TrainExList, ClassDist, ClassTupleL, ClassIndex, I, DistL, CatL),!,
	(member(cat_split(all), Param) ->
	  i_gain(DistL, ClassDist, NGain),
	  NCatL = CatL
	;
	  %trace,
	  best_binary_i_gain(DistL, ClassDist, NGain, CatL, NCatL)
	),
	(NGain = first ->
	 Gain = 0
	;
	 Gain = NGain
	),
	%atom_codes(Op,[61]), %61 is =
	%Split=[LS,RS],
	%last(LS, Val1-(_,_)),
	%RS=[Val2-(_,_)|_],
	%(Val1 = Val2 ->    %Do not split into impossible values!
	%   FGain = 0
	%;
	%   FGain = Gain
				%),
 %       write('Gain: '), write(Gain),nl,
%	(Gain = 0.008425543864907459 -> trace;true),
        %Split=[LS,RS],
        %length(LS,LenLs),
        %length(RS,LenRs), 
        %write('Left: '),write(LenLs),nl, 
        %write('Right: '),write(LenRs),nl, 
%	write('I = '),write(I),nl,
	%(I = 3 -> trace; true),!,
	(TGain > NGain ->
	  !,select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TF-TSplit-TGain, FSG)
	;
	  !,select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, F-NCatL-Gain, FSG)
	).
select_feature([I-F|FeatureDTList], TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, first-_-_, FSG):-
	F = numeric(_),!, %skall vara numeric(F)!?
	%write('Innan numeric split...'),nl,
	%write('ClassDist: '),write(ClassDist),nl,
				%write('TrainExList: '), write(TrainExList),
	%trace,
        split(TrainExList, ClassDist, _, ClassIndex, I, F, Split-NGain),
	(NGain = first ->
	 Gain = 0
	;
	 Gain = NGain
	),
	Split=[LS,RS],
	(LS = [] ; RS = [] ->   % ignore empty sub_nodes...
	   FGain = 0
	;
	   last(LS, Val1-(_,_)),
	   RS=[Val2-(_,_)|_]
	),
	(Val1 = Val2 ->    %Do not split into impossible values!
	   FGain = 0
	;
	   FGain = Gain
	),	 
        %write('Gain: '), write(Gain),nl,
        Split=[LS,RS],
        %length(LS,LenLs),
        %length(RS,LenRs), 
        %write('Left: '),write(LenLs),nl, 
        %write('Right: '),write(LenRs),nl, 
	%write('I = '),write(I),nl,
	%(I = 3 -> trace; true),!,
	select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, F-Split-Gain, FSG).
select_feature([I-F|FeatureDTList], TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TF-TSplit-TGain, FSG):-
	F = numeric(_),!, %skall vara numeric(F)!?
	%write('Innan numeric split...'),nl,
	%write('ClassDist: '),write(ClassDist),nl,
				%write('TrainExList: '), write(TrainExList),
	%trace,
        split(TrainExList, ClassDist, _, ClassIndex, I, F, Split-NGain),
	(NGain = first ->
	 Gain = 0
	;
	 Gain = NGain
	),
	Split=[LS,RS],
	(LS = [] ; RS = [] ->   % ignore empty sub_nodes...
	   FGain = 0
	;
	   last(LS, Val1-(_,_)),
	   RS=[Val2-(_,_)|_]
	),
	(Val1 = Val2 ->    %Do not split into impossible values!
	   FGain = 0
	;
	   FGain = Gain
	),
	%Split=[LS,RS],
	%last(LS, Val1-(_,_)),
	%RS=[Val2-(_,_)|_],
	%(Val1 = Val2 ->    %Do not split into impossible values!
	%   FGain = 0
	%;
	%   FGain = Gain
	%),	 
        %write('Gain: '), write(Gain),nl,
        Split=[LS,RS],
        %length(LS,LenLs),
        %length(RS,LenRs), 
        %write('Left: '),write(LenLs),nl, 
        %write('Right: '),write(LenRs),nl, 
	%write('I = '),write(I),nl,
	%(I = 3 -> trace; true),!,
	(TGain > Gain ->
	  !,select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, TF-TSplit-TGain, FSG)
	;
          !,select_feature(FeatureDTList, TrainExList, ClassTupleL, ClassDist, ClassIndex, Param, F-Split-Gain, FSG)
	).

cat_split(_, [], _, _, _, _, [], []). %:-
	%write('Tree: Out of Examples').
cat_split([], _, _, _, _, _, [], []).
cat_split([Cat|Cats], TrainExList, ClassDist, ClassTupleL, ClassIndex, I, [CatClassDist|RestDist], [CatSplitInfo|RestGains]):-
	copy_term(ClassTupleL, CClassTupleL),
	get_count_list(TrainExList, Cat, I, ClassIndex, CClassTupleL, CatClassDist, CatSplitInfo),
	cat_split(Cats, TrainExList, ClassDist, ClassTupleL, ClassIndex, I, RestDist, RestGains).

get_count_list([], _, _, _, InsDist, InsDist, _-[InsDist-[]]).
get_count_list([Ex|Exs], Cat, I, ClassIndex, TClassCountL, FinalCountL, Cat-[InsDist-[Ex|Inside]]):-
	arg(I, Ex, Cat),!,
	arg(ClassIndex, Ex, Class),
	select(No-Class, TClassCountL, RestCountL),
	NewNo is No + 1,
	get_count_list(Exs, Cat, I, ClassIndex, [NewNo-Class|RestCountL], FinalCountL, Cat-[InsDist-Inside]).
get_count_list([_|Exs], Cat, I, CI, TClassCountL, FinalCountL, Cat-[InsDist-Inside]):-
	get_count_list(Exs, Cat, I, CI, TClassCountL, FinalCountL, Cat-[InsDist-Inside]).

split([], _, _, _, _, _, []). %:- write('Out of Examples').
%split(_, _, _, _, _, class(_), []-0). %:-write('Using classlabel - remove').
split(Exs, ClassDist, ClassTupleL, ClassIndex, I, numeric(_), Split-Gain):-
	get_all_values(Exs, ClassIndex, 1, I, KeyValue),
        %length(KeyValue,KVL),
        %write('KVL: '),write(KVL),nl,
	keysort(KeyValue, Sorted),
        %length(Sorted,SKVL),
        %write('SortedKVL: '),write(SKVL),nl,
        %trace,  
	find_best_binary_split(Sorted, [], _, ClassDist, I, [[],[]], ClassTupleL-first, Split-Gain).

find_best_binary_split([], _, _, _, _, _, Split-Value, Split-Value).
find_best_binary_split([Arg-(Key,Class)|Sorted], _, _, ClassDist, I, [OldLC,_], _-first, Split-Value):- 
	find_change([Arg-(Key,Class)|Sorted], NewClass, Class, [0-Class], LeftC, Arg, LSplit, RSplit),
        %select(_N-Class,ClassTuplesL,Rest),
        %append(LeftC,Rest,CompLeftC), 
	merge_count(ClassDist, OldLC, LeftC, NewLeftC),
        right_count(NewLeftC, ClassDist, RightC),
        %length(LSplit,NLCLen),
        %length(RSplit,RCLen),
        %TotCount is NLCLen + RCLen,
        %write('Temp Tot Len(1): '), write(TotCoun),nl,  
	i_gain([NewLeftC, RightC], ClassDist, TValue),!,
	find_best_binary_split(RSplit, LSplit, NewClass, ClassDist, I, [NewLeftC, RightC], [LSplit, RSplit]-TValue, Split-Value).
find_best_binary_split([Arg-(Key,C)|Exs], OldLSplit, Class, ClassDist, I, [OldLC, _OldRC], [OLSplit, ORSplit]-TempV, Split-Value):-
	find_change([Arg-(Key,C)|Exs], NewClass, Class, [0-Class], LeftC, Arg, LSplit, RSplit),
	merge_count(ClassDist, OldLC, LeftC, NewLeftC),
	append(OldLSplit,LSplit,NewLSplit),
	right_count(NewLeftC, ClassDist, RightC),
        %length(NewLSplit,NLCLen),
        %length(RSplit,RCLen),
        %TotCount is NLCLen + RCLen,
        %write('Temp Tot Len(2): '),write(TotCount),nl,  
        %trace,
	i_gain([NewLeftC, RightC], ClassDist, TValue),!,
	(TValue > TempV ->
	  find_best_binary_split(RSplit, NewLSplit, NewClass, ClassDist, I, [NewLeftC, RightC], [NewLSplit, RSplit]-TValue, Split-Value)
	;
	  find_best_binary_split(RSplit, NewLSplit, NewClass, ClassDist, I, [NewLeftC, RightC], [OLSplit, ORSplit]-TempV, Split-Value)
	).

merge_count([], _OldLC, _NewLC, []).
merge_count([_-Class|Classes], OldLC, NewLC, [NewestC-Class|NewestLCs]):-
	member(OldC-Class,OldLC),
	member(NewC-Class,NewLC),!,
	NewestC is OldC + NewC,
	merge_count(Classes, OldLC, NewLC, NewestLCs).
merge_count([_-Class|Classes], OldLC, NewLC, [OldC-Class|NewestLCs]):-
	member(OldC-Class,OldLC),!,
	merge_count(Classes, OldLC, NewLC, NewestLCs).
merge_count([_-Class|Classes], OldLC, NewLC, [NewC-Class|NewestLCs]):-
	member(NewC-Class,NewLC),!,
	merge_count(Classes, OldLC, NewLC, NewestLCs).
merge_count([_|Classes], OldLC, NewLC, NewestLCs):-
	merge_count(Classes, OldLC, NewLC, NewestLCs).

right_count([], ClassDist, ClassDist).
right_count([LCount-Class|LCs], ClassDist, [RCount-Class|RCs]):-
        select(TCount-Class, ClassDist, NewClassDist),
	RCount is TCount - LCount,
      	right_count(LCs, NewClassDist, RCs).

%Add check for binary category, then we only need to do on i_gain test... i find_best?!
best_binary_i_gain(DistL, ClassDist, Gain, CatL, NCatL):-
        find_best_gain(DistL, [], ClassDist, 0, _, _, Gain, PartDist, RestDist),
        select(F-[PartDist-Ex], CatL, RestCatL),
	get_rest_ex(RestCatL, RestEx),
	NCatL = [F-[PartDist-Ex], not(F)-[RestDist-RestEx]].

get_rest_ex([], []).
get_rest_ex([_-[_-Ex]|Rest], [Ex|RestEx]):-
        get_rest_ex(Rest, RestEx).

find_best_gain([], _, _, Gain, Dist, RestDist, Gain, Dist, RestDist).	
find_best_gain([PartDist|RestDist], LeftParts, ClassDist, TempG, TempDist, TempRestDist, Gain, Dist, FRestDist):-
	merge_cat_count(RestDist, MergedRestDist),
	merge_cat_count(LeftParts, LeftDist),
	join([MergedRestDist], [LeftDist], AppDist),
	merge_cat_count(AppDist, TotRestDist),
	i_gain([PartDist,TotRestDist], ClassDist, CurrGain),!,
	(TempG < CurrGain ->
	  NewTempG = CurrGain,
	  NewTempDist = PartDist,
	  NewTempRestDist = TotRestDist,
	  find_best_gain(RestDist, [PartDist|LeftParts], ClassDist, NewTempG, NewTempDist, NewTempRestDist, Gain, Dist, FRestDist)
	;
	  find_best_gain(RestDist, [PartDist|LeftParts], ClassDist, TempG, TempDist, TempRestDist, Gain, Dist, FRestDist)
	).

%Remove all empty list in list keep only filled ones..
join([[]], [[]], []).
join([[]], L, L).
join(L, [[]], L).
join([F], [N], [F,N]).

merge_cat_count([], []).
merge_cat_count([FirstD|RD], Dist):-
	merge_cat_count(RD, FirstD, Dist).

merge_cat_count([], Dist, Dist).
merge_cat_count(Dist, [], Dist).
merge_cat_count([SubL|R], TempD, Dist):-
       m_count(SubL, TempD, NewTempD),
       merge_cat_count(R, NewTempD, Dist).

m_count([], Dist, Dist).
m_count([No-Class|R], TempD, Dist):-
	select(TempNo-Class, TempD, RestD),
        NewTempNo is TempNo + No,
	select(NewTempNo-Class, NewTempD, RestD),
	m_count(R, NewTempD, Dist).

i_gain(CountList, ClassDist, Gain):-
        %laplace_corr(ClassDist, NewClassDist, CountList, NewCountList),  %chec this one later one
	sum(ClassDist, 0, Tot),
        %sum_l(CountList, PartTot),
        log_2_mult(Multiple),
	i(ClassDist, Tot, Tot, Multiple, 0, IVal),
	i(CountList, Tot, Multiple, 0, NewIVal),!,
	Gain is IVal -  NewIVal.
/*
laplace_corr(ClassDist, NewClassDist, CountList, NewCountList):-
    length(CountList, L),
    add_size(ClassDist, L, NewClassDist),
    add_size(CountList, 1, NewCountList).

add_size([], _, []).
add_Size([Count-Class|CC], Size, [NCount-Class|NCC]):-
    NCount is Count + Size,
    add_size(CC, NCC).
*/
%i/5 - only for features
i([], _, _, FVal, FVal).
i([L|R], Tot, Mul, TVal, FVal):-
	%is_list(L),!,   
        sum(L, 0, Size),
	i(L, Tot, Size,  Mul, TVal, FVal1),
        i(R, Tot, Mul, 0, FVal2),!,
        FVal is FVal1 + FVal2.

%i/6 - only for tot, (and called by i/5)
i([], Tot, Size, _, Entropy, FVal):-
    FVal is Entropy * Size / Tot,!.
i([0-_|R], Tot, Size, Mul, TVal, FVal):-
        i(R, Tot, Size, Mul, TVal, FVal).
i([E-_|R], Tot, Size, Mul, TVal, FVal):-
	Quota is E / Size,
        (Quota < 0 ->
	   trace, true
	 ;
	   NTVal is TVal + -(Quota * log(Quota) * Mul),
	   i(R, Tot, Size, Mul, NTVal, FVal)
         ).

sum_l([],0):-!.
sum_l([L|Ls],T):-
    sum(L, 0, T1),
    sum_l(Ls,T2),
    T is T1 + T2.

sum([],Tot,Tot):-!.
sum([Count-_|Cs],TTot,FTot):-
	NewTTot is TTot + Count,
	sum(Cs, NewTTot,FTot).

find_change([], Class, Class, FCount, FCount, _, [], []).
find_change([Arg-(Key,Class)|AKCs], NewClass, Class, TempCount, FCount, _, [Arg-(Key,Class)|Rest], RSplit):- %Detect change
        select(TempC-C,TempCount,RTempC),
	NewTCount is TempC + 1,
        append([NewTCount-C], RTempC, NewTempCount),!,
	find_change(AKCs, NewClass, Class, NewTempCount, FCount, Arg, Rest, RSplit).
find_change([Arg-(Key,C)|AKCs], NewClass, Class, TempCount, FCount, Arg, [Arg-(Key,C)|Rest], RSplit):- %Fetch equal values
    (select(TempC-C, TempCount, RTempC) ->	
        NewTCount is TempC + 1,!,
        append([NewTCount-C], RTempC, NewTempCount),
    	find_end(AKCs, NewClass, Class, NewTempCount, FCount, Arg, Rest, RSplit) %look for end
    ;    
        append([1-C], TempCount, NewTempCount),
	find_end(AKCs, NewClass, Class, NewTempCount, FCount, Arg, Rest, RSplit)
    ). 
find_change([Arg-(Key,NewClass)|AKCs], NewClass, _Class, FCount, FCount, _, [],[Arg-(Key,NewClass)|AKCs]). %end
find_end([Arg-(Key,C)|AKCs], NewClass, Class, TempCount, FCount, Arg, [Arg-(Key,C)|Rest], RSplit):- %Fetch equal values
    (select(TempC-C, TempCount, RTempC) ->	
        NewTCount is TempC + 1,!,
        append([NewTCount-C], RTempC, NewTempCount),
    	find_end(AKCs, NewClass, Class, NewTempCount, FCount, Arg, Rest, RSplit) %look for end
    ;    
        append([1-C], TempCount, NewTempCount),
	find_end(AKCs, NewClass, Class, NewTempCount, FCount, Arg, Rest, RSplit)
     ).
find_end([Arg-(Key,NewClass)|AKCs], NewClass, _Class, FCount, FCount, _, [],[Arg-(Key,NewClass)|AKCs]). %end

get_all_values([],_, _Key,_I,[]).
get_all_values([Ex|Exs], ClassIndex, Key, I, KeyVals):-  %ignore '?' missing values
	arg(I,Ex,'?'),!,
	get_all_values(Exs, ClassIndex, Key, I, KeyVals).
get_all_values([Ex|Exs], ClassIndex, Key, I, [Arg-(Key,Class)|KeyVals]):-
	arg(I,Ex,Arg),
	arg(ClassIndex,Ex,Class),
	NewKey is Key + 1,
	get_all_values(Exs, ClassIndex, NewKey, I, KeyVals).

count_classes(TrainEx, Key, FeatureDTList, I, ClassTuplesL, ClassTuplesList):-  
    findall(C,Key:class(C),ClassesL),
    class_tuple_c(ClassesL, ClassTuplesL), 
    nth1(I, FeatureDTList, class(_)),
    do_count_classes(TrainEx, I, ClassTuplesL, ClassTuplesList).

create_class_tuples(Key, ClassTuplesL):-
    findall(C,Key:class(C),ClassesL),
    class_tuple_c(ClassesL, ClassTuplesL).

class_tuple_c([],[]).
class_tuple_c([Class|Classes],[0-Class|Tuples]):-
	class_tuple_c(Classes,Tuples).

do_count_classes([], _, CTL, CTL).
do_count_classes([Ex|ExList], I, CTL, ClassList):-
	arg(I,Ex,C),
	select(Count-C,CTL,CTLPart),
	NewCount is Count + 1,
	do_count_classes(ExList, I, [NewCount-C|CTLPart], ClassList).
















